// @see http://javascript.ru/php/sprintf
function ETHEREUM_WALLET_sprintf( ) {	// Return a formatted string
	// 
	// +   original by: Ash Searle (http://hexmen.com/blog/)
	// + namespaced by: Michael White (http://crestidg.com)

	var regex = /%%|%(\d+\$)?([-+#0 ]*)(\*\d+\$|\*|\d+)?(\.(\*\d+\$|\*|\d+))?([scboxXuidfegEG])/g;
	var a = arguments, i = 0, format = a[i++];

	// pad()
	var pad = function(str, len, chr, leftJustify) {
		var padding = (str.length >= len) ? '' : Array(1 + len - str.length >>> 0).join(chr);
		return leftJustify ? str + padding : padding + str;
	};

	// justify()
	var justify = function(value, prefix, leftJustify, minWidth, zeroPad) {
		var diff = minWidth - value.length;
		if (diff > 0) {
			if (leftJustify || !zeroPad) {
			value = pad(value, minWidth, ' ', leftJustify);
			} else {
			value = value.slice(0, prefix.length) + pad('', diff, '0', true) + value.slice(prefix.length);
			}
		}
		return value;
	};

	// formatBaseX()
	var formatBaseX = function(value, base, prefix, leftJustify, minWidth, precision, zeroPad) {
		// Note: casts negative numbers to positive ones
		var number = value >>> 0;
		prefix = prefix && number && {'2': '0b', '8': '0', '16': '0x'}[base] || '';
		value = prefix + pad(number.toString(base), precision || 0, '0', false);
		return justify(value, prefix, leftJustify, minWidth, zeroPad);
	};

	// formatString()
	var formatString = function(value, leftJustify, minWidth, precision, zeroPad) {
		if (precision != null) {
			value = value.slice(0, precision);
		}
		return justify(value, '', leftJustify, minWidth, zeroPad);
	};

	// finalFormat()
	var doFormat = function(substring, valueIndex, flags, minWidth, _, precision, type) {
		if (substring == '%%') return '%';

		// parse flags
		var leftJustify = false, positivePrefix = '', zeroPad = false, prefixBaseX = false;
		for (var j = 0; flags && j < flags.length; j++) switch (flags.charAt(j)) {
			case ' ': positivePrefix = ' '; break;
			case '+': positivePrefix = '+'; break;
			case '-': leftJustify = true; break;
			case '0': zeroPad = true; break;
			case '#': prefixBaseX = true; break;
		}

		// parameters may be null, undefined, empty-string or real valued
		// we want to ignore null, undefined and empty-string values
		if (!minWidth) {
			minWidth = 0;
		} else if (minWidth == '*') {
			minWidth = +a[i++];
		} else if (minWidth.charAt(0) == '*') {
			minWidth = +a[minWidth.slice(1, -1)];
		} else {
			minWidth = +minWidth;
		}

		// Note: undocumented perl feature:
		if (minWidth < 0) {
			minWidth = -minWidth;
			leftJustify = true;
		}

		if (!isFinite(minWidth)) {
			throw new Error('sprintf: (minimum-)width must be finite');
		}

		if (!precision) {
			precision = 'fFeE'.indexOf(type) > -1 ? 6 : (type == 'd') ? 0 : void(0);
		} else if (precision == '*') {
			precision = +a[i++];
		} else if (precision.charAt(0) == '*') {
			precision = +a[precision.slice(1, -1)];
		} else {
			precision = +precision;
		}

		// grab value using valueIndex if required?
		var value = valueIndex ? a[valueIndex.slice(0, -1)] : a[i++];

		switch (type) {
			case 's': return formatString(String(value), leftJustify, minWidth, precision, zeroPad);
			case 'c': return formatString(String.fromCharCode(+value), leftJustify, minWidth, precision, zeroPad);
			case 'b': return formatBaseX(value, 2, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
			case 'o': return formatBaseX(value, 8, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
			case 'x': return formatBaseX(value, 16, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
			case 'X': return formatBaseX(value, 16, prefixBaseX, leftJustify, minWidth, precision, zeroPad).toUpperCase();
			case 'u': return formatBaseX(value, 10, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
			case 'i':
			case 'd': {
						var number = parseInt(+value);
						var prefix = number < 0 ? '-' : positivePrefix;
						value = prefix + pad(String(Math.abs(number)), precision, '0', false);
						return justify(value, prefix, leftJustify, minWidth, zeroPad);
					}
			case 'e':
			case 'E':
			case 'f':
			case 'F':
			case 'g':
			case 'G':
						{
						var number = +value;
						var prefix = number < 0 ? '-' : positivePrefix;
						var method = ['toExponential', 'toFixed', 'toPrecision']['efg'.indexOf(type.toLowerCase())];
						var textTransform = ['toString', 'toUpperCase']['eEfFgG'.indexOf(type) % 2];
						value = prefix + Math.abs(number)[method](precision);
						return justify(value, prefix, leftJustify, minWidth, zeroPad)[textTransform]();
					}
			default: return substring;
		}
	};

	return format.replace(regex, doFormat);
}

// http://locutus.io/php/math/round/
function ETHEREUM_WALLET_round (value, precision, mode) {
  //  discuss at: http://locutus.io/php/round/
  // original by: Philip Peterson
  //  revised by: Onno Marsman (https://twitter.com/onnomarsman)
  //  revised by: T.Wild
  //  revised by: Rafał Kukawski (http://blog.kukawski.pl)
  //    input by: Greenseed
  //    input by: meo
  //    input by: William
  //    input by: Josep Sanz (http://www.ws3.es/)
  // bugfixed by: Brett Zamir (http://brett-zamir.me)
  //      note 1: Great work. Ideas for improvement:
  //      note 1: - code more compliant with developer guidelines
  //      note 1: - for implementing PHP constant arguments look at
  //      note 1: the pathinfo() function, it offers the greatest
  //      note 1: flexibility & compatibility possible
  //   example 1: round(1241757, -3)
  //   returns 1: 1242000
  //   example 2: round(3.6)
  //   returns 2: 4
  //   example 3: round(2.835, 2)
  //   returns 3: 2.84
  //   example 4: round(1.1749999999999, 2)
  //   returns 4: 1.17
  //   example 5: round(58551.799999999996, 2)
  //   returns 5: 58551.8

  var m, f, isHalf, sgn // helper variables
  // making sure precision is integer
  precision |= 0
  m = Math.pow(10, precision)
  value *= m
  // sign of the number
  sgn = (value > 0) | -(value < 0)
  isHalf = value % 1 === 0.5 * sgn
  f = Math.floor(value)

  if (isHalf) {
    switch (mode) {
      case 'PHP_ROUND_HALF_DOWN':
      // rounds .5 toward zero
        value = f + (sgn < 0)
        break
      case 'PHP_ROUND_HALF_EVEN':
      // rouds .5 towards the next even integer
        value = f + (f % 2 * sgn)
        break
      case 'PHP_ROUND_HALF_ODD':
      // rounds .5 towards the next odd integer
        value = f + !(f % 2)
        break
      default:
      // rounds .5 away from zero
        value = f + (sgn > 0)
    }
  }

  return (isHalf ? value : Math.round(value)) / m
}

if (!String.prototype.startsWith) {
  Object.defineProperty(String.prototype, 'startsWith', {
    enumerable: false,
    configurable: false,
    writable: false,
    value: function(searchString, position) {
      position = position || 0;
      return this.indexOf(searchString, position) === position;
    }
  });
}

// From https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/keys
if (!Object.keys) {
  Object.keys = (function() {
    'use strict';
    var hasOwnProperty = Object.prototype.hasOwnProperty,
        hasDontEnumBug = !({ toString: null }).propertyIsEnumerable('toString'),
        dontEnums = [
          'toString',
          'toLocaleString',
          'valueOf',
          'hasOwnProperty',
          'isPrototypeOf',
          'propertyIsEnumerable',
          'constructor'
        ],
        dontEnumsLength = dontEnums.length;

    return function(obj) {
      if (typeof obj !== 'object' && (typeof obj !== 'function' || obj === null)) {
        throw new TypeError('Object.keys called on non-object');
      }

      var result = [], prop, i;

      for (prop in obj) {
        if (hasOwnProperty.call(obj, prop)) {
          result.push(prop);
        }
      }

      if (hasDontEnumBug) {
        for (i = 0; i < dontEnumsLength; i++) {
          if (hasOwnProperty.call(obj, dontEnums[i])) {
            result.push(dontEnums[i]);
          }
        }
      }
      return result;
    };
  }());
}

function ETHEREUM_WALLET_addressCopied(e) {
    console.info('Action:', e.action);
    console.info('Text:', e.text);
    console.info('Trigger:', e.trigger);
    
    alert(window.ethereumWallet.str_copied_msg);

    e.clearSelection();
}

//function ETHEREUM_WALLET_copyAddress(e) {
//	e.preventDefault();
//	// copy in any case
//	var $temp = jQuery("<input>");
//	jQuery("body").append($temp);
//
//    var target = e.target;
//    if ("BUTTON" !== e.target.tagName) {
//        target = target.parentElement;
//    }
//	var id = jQuery(target).data("input-id");
//	console.log("Copy from: ", id);
//
//	var value = jQuery("#" + id).val();
//	console.log("Value to copy: ", value);
//
//	$temp.val(value).select();		
//	document.execCommand("copy");
//	$temp.remove();
//
//    alert(window.ethereumWallet.str_copied_msg);
//}

function ETHEREUM_WALLET_init() {
	if ("undefined" === typeof window['ethereumWallet']) {
        return;
    }
	if (window.ethereumWallet.initialized === true) {
        return;
    }
	if ("undefined" !== typeof window.ethereumWallet['web3Endpoint']) {
		if (typeof window !== 'undefined' && typeof window.web3 !== 'undefined') {
            var injectedProvider = window.web3.currentProvider;
            window.ethereumWallet.web3metamask = new Web3(injectedProvider)
		}
        window.ethereumWallet.web3 = new Web3(new Web3.providers.HttpProvider(window.ethereumWallet.web3Endpoint));
	}
//    jQuery(".ethereum-wallet-copy-button").click(ETHEREUM_WALLET_copyAddress);
    // Init QR codes
    window.ethereumWallet.copytoclipboardbuttons = [];
    for (var i = 1; i < 100; i++) {
        if (jQuery('.ethereum-wallet-account-canvas-qr' + i).length > 0) {
            jQuery('.ethereum-wallet-account-canvas-qr' + i).qrcode({
                text : jQuery('#ethereum-wallet-account' + i).val()
            });
            var clipboard = new ClipboardJS('#ethereum-wallet-account-copy-qr-button' + i);
            clipboard.on("success", ETHEREUM_WALLET_addressCopied);
            window.ethereumWallet.copytoclipboardbuttons.push(clipboard);
        }
        if (jQuery('.ethereum-wallet-account-management-export-canvas-qr' + i).length > 0) {
            jQuery('.ethereum-wallet-account-management-export-canvas-qr' + i).qrcode({
                text : jQuery('#ethereum-wallet-account-management-export' + i).val()
            });
            var clipboard = new ClipboardJS('#ethereum-wallet-account-management-export-copy-qr-button' + i);
            clipboard.on("success", ETHEREUM_WALLET_addressCopied);
            window.ethereumWallet.copytoclipboardbuttons.push(clipboard);
        }
    }
    if (jQuery('.ethereum-wallet-history-table').length > 0) {
        jQuery('.ethereum-wallet-history-table').DataTable({
            language: {
                url: window.ethereumWallet.localePath
            },
            columns: [
                { data: "counter" },
                { 
                    data: "value",
                    type: "num",
                    render: function ( value, type, row ) {
                        if ("sort" === type) {
                            if (!row.dir) {
                                return -parseFloat(value);
                            }
                            return parseFloat(value);
                        }
                        if (!row.dir) {
                            return '<span class="ethereum-wallet-history-dir ethereum-wallet-history-minus">-</span>' + value;
                        }
                        return value;
                    }
                },
                { 
                    data: "timeStamp",
                    render: function ( timeStamp, type, row ) {
                        if ("sort" === type) {
                            return timeStamp;
                        }
                        if ('undefined' !== typeof row.unconfirmed && row.unconfirmed) {
                            return window.ethereumWallet.str_tx_pending_msg;
                        }
                        var days = (new Date - new Date(timeStamp * 1000)) / (24 * 3600 * 1000);
                        var dateString = '';
                        if (days >= 1) {
                            dateString = ETHEREUM_WALLET_sprintf(
                                window.ethereumWallet.str_date_days_fmt_msg
                                , Math.floor(days)
                            );
                        } else {
                            var hours = 24 * days;
                            if (hours >= 1) {
                                dateString = ETHEREUM_WALLET_sprintf(
                                    window.ethereumWallet.str_date_hours_fmt_msg
                                    , Math.floor(hours)
                                );
                            } else {
                                var minutes = 60 * hours;
                                if (minutes >= 1) {
                                    dateString = ETHEREUM_WALLET_sprintf(
                                        window.ethereumWallet.str_date_minutes_fmt_msg
                                        , Math.floor(minutes)
                                    );
                                } else {
                                    dateString = window.ethereumWallet.str_date_recently_msg;
                                }
                            }
                        }
                        return dateString;
                    }
                },
                { 
                    data: "hash",
                    render: function ( hash, type, row ) {
                        var thash = hash.substr(0, 8);
                        var blockchain_network = '';
                        if ('mainnet' !== window.ethereumWallet.blockchain_network) {
                            blockchain_network = window.ethereumWallet.blockchain_network + '.';
                        }
                        return '<a target="_blank" href="https://' + blockchain_network + 'etherscan.io/tx/' + hash + '">' + thash + '</a>';
                    }
                }
            ]
        });
    }
    if (jQuery('.ethereum-wallet-accounts-table').length > 0) {
        var colsCount = jQuery('table.ethereum-wallet-accounts-table > thead > tr').children().length;
        var wpnonce = jQuery('table.ethereum-wallet-accounts-table').data('wpnonce');
        // Rank | 
        // User profile image (VIA BUDDYPRESS) | 
        // Username (VIA BUDDYPRESS) | 
        // Wallet ETH Address | 
        // ETH FIAT Balance in Wallet (Example: $357.89) | 
        // TX Count
        var columns = [];
        columns.push({ 
            data: "rank",
            orderable: false,
            type: "num"
        });
        if (6 === colsCount) {
            columns.push({ 
                data: "avatarUrl",
                orderable: false,
                render: function ( value, type, row ) {
                    return '<img class="ethereum-wallet-accounts-table-avatar-img" src="' + value + '">';
                }
            });
            columns.push({ 
                data: "username",
                orderable: false,
            });
        }
        columns.push({ 
            data: "ether_address",
            orderable: false,
            render: function ( hash, type, row ) {
                var thash = hash.substr(0, 8);
                var blockchain_network = '';
                if ('mainnet' !== window.ethereumWallet.blockchain_network) {
                    blockchain_network = window.ethereumWallet.blockchain_network + '.';
                }
                return '<a target="_blank" href="https://' + blockchain_network + 'etherscan.io/address/' + hash + '" title="' + thash + '">' + thash + '</a>';
            }
        });
        columns.push({ 
            data: "fiat_balance",
            orderable: false,
        });
        columns.push({ 
            data: "tx_count",
            orderable: false,
        });
        jQuery('.ethereum-wallet-accounts-table').DataTable({
            language: {
                url: window.ethereumWallet.localePath
            },
            order: [],
            processing: true,
            serverSide: true,
            ajax: {
                url: "/wp-json/ethereumwallet/v1/accounts",
                type: "POST",
//                dataSrc: "",
                data: function ( d ) {
                    d._wpnonce = wpnonce;
                    d.page = Math.floor(d.start / d.length) + 1;
                    d.per_page = d.length;
                    d.s = d.search.value;
                    // d.custom = $('#myInput').val();
                    // etc
                },
            },
            columns: columns
        });
    }
//    jQuery(".ethereum-wallet-copy-button").click(ETHEREUM_WALLET_copyAddress);
    ETHEREUM_WALLET_update_transactions(null, function(err, updated) {
        if (err) {
            ETHEREUM_WALLET_hide_wait_icon();
            console.log(err);
//            alert(err);
            return;
        }

        var transactionHash = window.ethereumWallet.user_wallet_last_txhash;
        if ('' !== transactionHash) {
            var blockCount = 1; // TODO: add admin setting
            var timeout = 5 * 60; // 5 minutes in seconds
            var transactionTimeStr = window.ethereumWallet.user_wallet_last_txtime;
            // for old transactions made before this update
            // set it to expire immediately
            var transactionTime = timeout * 1000;
            if ('' !== transactionTimeStr) {
                // for new transactions count it's real time
                transactionTime = 1000 * parseInt(transactionTimeStr);
            }
            var tm = new Date;
            var startTime = new Date(tm.getTime() - transactionTime);

            if ((tm - startTime) < 1000 * timeout) {
                ETHEREUM_WALLET_show_wait_icon();
                ETHEREUM_WALLET_awaitBlockConsensus(transactionHash, blockCount, timeout, transactionTime, function(err, transaction_receipt) {
                    if (err) {
                        ETHEREUM_WALLET_hide_wait_icon();
                        console.log(err);
                        return;
                    }
                    function ETHEREUM_WALLET_update_transactions_aux_() {
                        ETHEREUM_WALLET_update_transactions(transactionHash, function(err, updated) {
                            if (err) {
                                ETHEREUM_WALLET_hide_wait_icon();
                                console.log(err);
//                                alert(err);
                                return;
                            }
                            if (updated) {
                                ETHEREUM_WALLET_hide_wait_icon();
                                return;
                            }
                            setTimeout(ETHEREUM_WALLET_update_transactions_aux_, 5000);
                        });
                    };
                    // wait 30 seconds for etherscan API to reflect tx changes
                    setTimeout(ETHEREUM_WALLET_update_transactions_aux_, 30000);

                    // update balance values
                    ETHEREUM_WALLET_update_balances();
                });
            }
        }
        window.ethereumWallet.initialized = true;
    });
    setInterval(ETHEREUM_WALLET_update_balances, 10000);
    
    ETHEREUM_WALLET_init_qr_scanner();
}

function ETHEREUM_WALLET_init_qr_scanner() {
    if (0 === jQuery('.ethereum-wallet-sendform-shortcode').length) {
        return;
    }
    var video = null;
    var canvasElement = document.getElementById("ethereum-wallet-to-canvas-qr1");
    var canvas = canvasElement.getContext("2d");
    var timeout = 250;
    var color_ = "#FF3B58";

    function drawLine(begin, end, color) {
        canvas.beginPath();
        canvas.moveTo(begin.x, begin.y);
        canvas.lineTo(end.x, end.y);
        canvas.lineWidth = 4;
        canvas.strokeStyle = color;
        canvas.stroke();
    }

    // This event is fired when a collapse element has been made visible to the user (will wait for CSS transitions to complete).
    jQuery('#ethereum-wallet-to-qr1').on('shown.bs.collapse', function () {
        // Use facingMode: environment to attemt to get the front camera on phones
        navigator.mediaDevices.getUserMedia({
            video: { facingMode: "environment"/*, width: 256, height: 256 */} 
        }).then(function(stream) {
            jQuery('#ethereum-wallet-to-canvas-qr1-loading-message').addClass("d-none");
            if (!video) {
                video = document.createElement("video");
                video.setAttribute("playsinline", true); // required to tell iOS safari we don't want fullscreen
            }
            video.srcObject = stream;
            video.onloadedmetadata = function(e) {
                video.play();
            };
            setTimeout(function() {
                requestAnimationFrame(tick);
            }, timeout);
        }).catch(function(err) {
            console.error(err.name + ": " + err.message);
            alert(err.name + ": " + err.message);
            jQuery('#ethereum-wallet-to-canvas-qr1-loading-message').removeClass("d-none");
        });
    });
    jQuery('#ethereum-wallet-to-qr1').on('hidden.bs.collapse', function () {
        if (video) {
            video.srcObject.getTracks().forEach(function(track) {
                track.stop();
            });
        }
    });

    function tick() {
        if (!video) {
            return;
        }
        if (video.readyState === video.HAVE_ENOUGH_DATA) {
            canvasElement.hidden = false;
            canvas.drawImage(video, 0, 0, canvasElement.width, canvasElement.height);
            var imageData = canvas.getImageData(0, 0, canvasElement.width, canvasElement.height);
            var code = jsQR(imageData.data, imageData.width, imageData.height, {
                inversionAttempts: "dontInvert",
            });
            if (code) {
                drawLine(code.location.topLeftCorner, code.location.topRightCorner, color_);
                drawLine(code.location.topRightCorner, code.location.bottomRightCorner, color_);
                drawLine(code.location.bottomRightCorner, code.location.bottomLeftCorner, color_);
                drawLine(code.location.bottomLeftCorner, code.location.topLeftCorner, color_);
                console.log(code.data);
                var data = code.data;
                // MetaMask QR-codes support
                var prefix = "ethereum:";
                if (data.startsWith(prefix)) {
                    data = data.substring(prefix.length, data.length);
                }
                jQuery('#ethereum-wallet-sendform-to').val(data);
            }
        }
        setTimeout(function() {
            requestAnimationFrame(tick);
        }, timeout);
    }
}

function ETHEREUM_WALLET_update_balances() {
    jQuery('.ethereum-wallet-balance-shortcode').each(function(){
        var $sc = jQuery(this);
        var $tokenDecimals = $sc.find('.ethereum-wallet-balance-token-decimals');
        var $tokenDecimalChar = $sc.find('.ethereum-wallet-balance-token-decimal-char');
        var $tokenSymbol = $sc.find('.ethereum-wallet-balance-token-symbol');
        var $balanceValue = $sc.find('.ethereum-wallet-balance-value');
        var tokenSymbol = $tokenSymbol.text();
        var tokenDecimals = parseInt($tokenDecimals.text());
        var tokenDecimalChar = $tokenDecimalChar.text();
        // ETH
        ETHEREUM_WALLET_get_eth_balance_by_account(window.ethereumWallet.user_wallet_address, function(error, balance) {
            if (error) {
                console.log(error);
                return;
            }
            var balanceWei = balance.toNumber() / Math.pow(10, 18);
            var balanceValue = '' + ETHEREUM_WALLET_round (balanceWei, tokenDecimals, "PHP_ROUND_HALF_UP");
            balanceValue = balanceValue.replace('.', tokenDecimalChar);
            if ($balanceValue.text() !== balanceValue) {
                // wait 30 seconds for etherscan API to reflect tx changes
                setTimeout(function(){
                    ETHEREUM_WALLET_update_transactions(null, function(err, updated) {
                        if (err) {
                            console.log(err);
                            return;
                        }
                    });
                }, 30000);
                if ('' === tokenSymbol) {
                    $balanceValue.html(balanceValue);
                } else {
                    $balanceValue.html(balanceValue + ' ' + tokenSymbol);
                }
            }
        });
    });
}

jQuery(document).ready(ETHEREUM_WALLET_init);
// proper init if loaded by ajax
jQuery(document).ajaxComplete(function( event, xhr, settings ) {
    // check if the loaded content contains our shortcodes
    if (!xhr ||
        'undefined' === typeof xhr.responseText || 
        (
            xhr.responseText.indexOf('ethereum-wallet-account-shortcode') === -1 &&
            xhr.responseText.indexOf('ethereum-wallet-sendform-shortcode') === -1 &&
            xhr.responseText.indexOf('ethereum-wallet-account-management-create-shortcode') === -1 &&
            xhr.responseText.indexOf('ethereum-wallet-account-management-select-shortcode') === -1 &&
            xhr.responseText.indexOf('ethereum-wallet-account-management-import-shortcode') === -1 &&
            xhr.responseText.indexOf('ethereum-wallet-account-management-export-shortcode') === -1 &&
            xhr.responseText.indexOf('ethereum-wallet-balance-shortcode') === -1 &&
            xhr.responseText.indexOf('ethereum-wallet-tokens-list-shortcode') === -1 &&
            xhr.responseText.indexOf('ethereum-wallet-history-shortcode') === -1
        )
    ) {
        return;
    }
    ETHEREUM_WALLET_init();
});

function ETHEREUM_WALLET_update_transactions(transactionHash, callback) {
    if (0 === jQuery('.ethereum-wallet-history-table').length) {
        setTimeout(function(){
            callback.call(null, null, true);
        }, 0);
        return;
    }
    ETHEREUM_WALLET_load_transactions(window.ethereumWallet.user_wallet_address, function(err, transactions) {
        if (err) {
            console.log(err);
            transactions = [];
        }

        ETHEREUM_WALLET_load_internal_transactions(window.ethereumWallet.user_wallet_address, function(err, internal_transactions) {
            if (err) {
                console.log(err);
                internal_transactions = [];
            }

            ETHEREUM_WALLET_get_unconfirmed_transactions(function(err, result) {
                if (err) {
                    console.log(err);
                    result = [];
                }
                result = result.concat(ETHEREUM_WALLET_prepare_transactions(transactions, window.ethereumWallet.user_wallet_address));
                var internal_result = ETHEREUM_WALLET_prepare_transactions(internal_transactions, window.ethereumWallet.user_wallet_address);
                result = result.concat(internal_result);
                // make unique
                var resultObj = {};
                result.forEach(function(r) {
                    resultObj[r.hash] = r;
                });
                var updated = 'undefined' !== typeof resultObj[transactionHash];
                result = Object.keys(resultObj).map(function(key) {
                    return resultObj[key];
                });
                result.sort(function(a, b) {
                    // DESC
                    return b.timeStamp - a.timeStamp;
                });
                // initialize the counter column
                result.forEach(function(t, i) {
                    t.counter = i + 1;
                });
                ETHEREUM_WALLET_render_transactions(result);
                callback.call(null, null, updated);
            });
        });
    });
}

function ETHEREUM_WALLET_render_transactions(transactions) {
    
    jQuery('.ethereum-wallet-history-table').each(function() {
        var $tbl = jQuery( this );
        var inputOnly = $tbl.hasClass('ethereum-wallet-history-table-direction-in');
        var outputOnly = $tbl.hasClass('ethereum-wallet-history-table-direction-out');
        
        var tbl = $tbl.DataTable();
        tbl.clear();
        tbl.rows.add(transactions.filter(function(t) {
            // t.dir == true means input
            if (t.dir && outputOnly) {
                return false;
            }
            if (!t.dir && inputOnly) {
                return false;
            }
            return true;
        }));
        tbl.draw(false);
    });
}

function ETHEREUM_WALLET_prepare_transactions(transactions, toAddress) {
    var result = [];
    for (var i = 0; i < transactions.length; i++) {
        var t = transactions[i];
        if (t.value === '0') {
            continue;
        }
        var value = parseFloat(t.value) / 1000000000000000000;

        result.push({
            value: value + ' ETH',
            timeStamp: parseInt(t.timeStamp),
            hash: t.hash,
            // true if input
            dir: toAddress.toLowerCase() === t.to.toLowerCase(),
        });
    }
    return result;
}

function ETHEREUM_WALLET_get_unconfirmed_transactions(callback) {
    var transactionHash = window.ethereumWallet.user_wallet_last_txhash;
    if ('' === transactionHash) {
        callback.call(null, null, []);
        return;
    }
    var transactionTimeStr = window.ethereumWallet.user_wallet_last_txtime;
    if ('' === transactionTimeStr) {
        callback.call(null, null, []);
        return;
    }
    var transactionTime = 1000 * parseInt(transactionTimeStr);
    var tm = new Date;
    var startTime = new Date(tm.getTime() - transactionTime);
    //var to = window.ethereumWallet.user_wallet_last_tx_to;
    var value = window.ethereumWallet.user_wallet_last_tx_value;
    var tokenSymbol = 'ETH';
    callback.call(null, null, [{
        value: value + ' ' + tokenSymbol,
        timeStamp: parseInt("" + Math.floor(startTime.getTime() / 1000)),
        // true if input
        dir: false,
        hash: transactionHash,
        unconfirmed: true
    }]);
    return;
}

function ETHEREUM_WALLET_validate_send_form() {
    var sTo = jQuery('#ethereum-wallet-sendform-to').val();
    if ("" === sTo) {
        return false;
    }
    var sAmount = jQuery('#ethereum-wallet-sendform-amount').val();
    if ("" === sAmount) {
        return false;
    }
    var gasLimit = 21000;
    var sCurrencyAddress = jQuery('#ethereum-wallet-sendform-currency').val();
    if (sCurrencyAddress !== "0x0000000000000000000000000000000000000001") {
        sAmount = "0";
        gasLimit = parseFloat(window.ethereumWallet.gasLimit);
    }
    var amountWei = parseFloat(sAmount) * Math.pow(10, 18);
    var minRequiredWei = gasLimit * parseFloat(window.ethereumWallet.gasPrice);
    var balanceWei = window.ethereumWallet.web3.eth.getBalance(window.ethereumWallet.user_wallet_address);
    if ((balanceWei.toNumber() - amountWei) < minRequiredWei) {
        alert(window.ethereumWallet.str_insufficient_eth_balance_msg);
        return false;
    }
}

function ETHEREUM_WALLET_validate_account_management_create_form() {
    var sName = jQuery('#ethereum-wallet-account-management-create-name').val();
    if ("" === sName) {
        return false;
    }
}

function ETHEREUM_WALLET_validate_account_management_import_form() {
    var sName = jQuery('#ethereum-wallet-account-management-import-name').val();
    if ("" === sName) {
        return false;
    }
    var sKey = jQuery('#ethereum-wallet-account-management-import-key').val();
    if ("" === sKey) {
        return false;
    }
}

function ETHEREUM_WALLET_validate_account_management_select_form() {
}

function ETHEREUM_WALLET_get_eth_balance_by_account(account, callback) {
    window.ethereumWallet.web3.eth.getBalance(account, function(error, balance) {
        if (error) {
            callback.call(null, error, null);
            return;
        }
        callback.call(null, null, balance);
    });
}

function ETHEREUM_WALLET_load_transactions(address, callback) {
    if ('' === address) {
        callback.call(null, "Empty address requested for ETHEREUM_WALLET_load_transactions!", null);
        return;
    }
    var blockchain_network = '';
    if ('mainnet' !== window.ethereumWallet.blockchain_network) {
        blockchain_network= '-' + window.ethereumWallet.blockchain_network;
    }
    // https://stackoverflow.com/a/42538992/4256005
    jQuery.ajax({
        headers:{  
            "Accept":"application/json",//depends on your api
            "Content-type":"application/x-www-form-urlencoded"//depends on your api
        },
        url: "https://api" + blockchain_network + ".etherscan.io/api?module=account&action=txlist&address=" + address + "&startblock=0&endblock=99999999&sort=desc",
        success:function(r) {
            if (r.status !== "1") {
                console.log(r.message);
                callback.call(null, r.message, null);
                return;
            }
            var trxns = r.result;
            callback.call(null, null, trxns);
        }
    });
}

function ETHEREUM_WALLET_load_internal_transactions(address, callback) {
    if ('' === address) {
        callback.call(null, "Empty address requested for ETHEREUM_WALLET_load_internal_transactions!", null);
        return;
    }
    var blockchain_network = '';
    if ('mainnet' !== window.ethereumWallet.blockchain_network) {
        blockchain_network= '-' + window.ethereumWallet.blockchain_network;
    }
    // https://stackoverflow.com/a/42538992/4256005
    jQuery.ajax({
        headers:{  
            "Accept":"application/json",//depends on your api
            "Content-type":"application/x-www-form-urlencoded"//depends on your api
        },
        url: "https://api" + blockchain_network + ".etherscan.io/api?module=account&action=txlistinternal&address=" + address + "&startblock=0&endblock=99999999&sort=desc",
        success:function(r) {
            if (r.status !== "1") {
                console.log(r.message);
                callback.call(null, r.message, null);
                return;
            }
            var trxns = r.result;
            callback.call(null, null, trxns);
        }
    });
}

// https://ethereum.stackexchange.com/a/2830
// @method ETHEREUM_WALLET_awaitBlockConsensus
// @param txhash is the transaction hash from when you submitted the transaction
// @param blockCount is the number of blocks to wait for.
// @param timeout in seconds 
// @param callback - callback(error, transaction_receipt) 
//
function ETHEREUM_WALLET_awaitBlockConsensus(txhash, blockCount, timeout, transactionTime, callback) {
    var txWeb3 = window.ethereumWallet.web3;
    var startBlock = Number.MAX_SAFE_INTEGER;
    var stateEnum = { start: 1, mined: 2, awaited: 3, confirmed: 4, unconfirmed: 5 };
    var savedTxInfo;
    // the actual transaction start time
    var startTime = new Date((new Date).getTime() - transactionTime);

    var pollState = stateEnum.start;
    
    var checkTimeout = function() {
        var tm = new Date;
        if ((tm - startTime) > 1000 * timeout) {
            pollState = stateEnum.unconfirmed;
            callback(new Error("Timed out, not confirmed"), null);
            return false;
        }
        return true;
    };

    var processPollStateStart = function() {
        if (!checkTimeout()) {
            return;
        }
        txWeb3.eth.getTransaction(txhash, function(e, txInfo) {
            if (e || txInfo === null) {
                console.log(e);
                setTimeout(processPollStateStart, 1000);
                return; // XXX silently drop errors
            }
            if (txInfo.blockHash !== null) {
                startBlock = txInfo.blockNumber;
                savedTxInfo = txInfo;
                console.log("mined");
                pollState = stateEnum.mined;
                setTimeout(processPollStateMined, 1);
            } else {
                setTimeout(processPollStateStart, 1000);
            }
        });
    };
    var processPollStateMined = function() {
        if (!checkTimeout()) {
            return;
        }
        txWeb3.eth.getBlockNumber(function (e, blockNum) {
            if (e) {
                console.log(e);
                setTimeout(processPollStateMined, 1000);
                return; // XXX silently drop errors
            }
            console.log("blockNum: ", blockNum);
            if (blockNum >= (blockCount + startBlock)) {
                pollState = stateEnum.awaited;
                setTimeout(processPollStateAwaited, 1);
            } else {
                console.log("Need to wait ", (blockCount + startBlock - blockNum), " blocks");
                setTimeout(processPollStateMined, 1000);
            }
        });
    };
    var processPollStateAwaited = function() {
        if (!checkTimeout()) {
            return;
        }
        txWeb3.eth.getTransactionReceipt(txhash, function(e, receipt) {
            if (e || receipt === null) {
                if (e) {
                    console.log("getTransactionReceipt: ", e);
                }
                setTimeout(processPollStateAwaited, 1000);
                return; // XXX silently drop errors.  TBD callback error?
            }
            console.log("receipt: ", receipt);
            // confirm we didn't run out of gas
            // XXX this is where we should be checking a plurality of nodes.  TBD
            if ('undefined' !== typeof receipt.status) {
                if (receipt.status === "0x1") {
                    pollState = stateEnum.confirmed;
                    callback(null, receipt);
                } else if (receipt.status === "0x0") {
                    if (receipt.gasUsed > savedTxInfo.gas) {
                        pollState = stateEnum.unconfirmed;
                        callback(new Error("we ran out of gas, not confirmed!"), null);
                    } else {
                        pollState = stateEnum.unconfirmed;
                        callback(new Error("bad tx status, not confirmed!"), null);
                    }
                } else {
                    // unknown status. pre-Byzantium
                    if (receipt.gasUsed >= savedTxInfo.gas) {
                        pollState = stateEnum.unconfirmed;
                        callback(new Error("we ran out of gas, not confirmed!"), null);
                    } else {
                        pollState = stateEnum.confirmed;
                        callback(null, receipt);
                    }
                }
            } else {
                // unknown status. pre-Byzantium
                if (receipt.gasUsed >= savedTxInfo.gas) {
                    pollState = stateEnum.unconfirmed;
                    callback(new Error("we ran out of gas, not confirmed!"), null);
                } else {
                    pollState = stateEnum.confirmed;
                    callback(null, receipt);
                }
            }
        });
    };

    processPollStateStart();
}

function ETHEREUM_WALLET_show_wait_icon() {
    jQuery('#ethereum-wallet-tx-in-progress-spinner').addClass('is-active');
    var transactionHash = window.ethereumWallet.user_wallet_last_txhash;
    if ('' !== transactionHash) {
        jQuery('#ethereum-wallet-tx-in-progress-alert').removeClass('hidden');
        jQuery('#ethereum-wallet-tx-in-progress-alert').removeAttr('hidden');
    }
}

function ETHEREUM_WALLET_hide_wait_icon() {
    jQuery('#ethereum-wallet-tx-in-progress-spinner').removeClass('is-active');
    jQuery('#ethereum-wallet-tx-in-progress-alert').addClass('hidden');
    jQuery('#ethereum-wallet-tx-in-progress-alert').attr('hidden', ' hidden');
}
