<?php
/*
Plugin Name: Ethereum Wallet
Plugin URI: https://wordpress.org/plugins/ethereum-wallet/
Description: Wallet for Ether and ERC20 tokens for WordPress
Version: 1.14.0
Author: ethereumicoio
Text Domain: ethereum-wallet
Domain Path: /languages
*/

if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

// Explicitly globalize to support bootstrapped WordPress
global 
	$ETHEREUM_WALLET_plugin_basename, $ETHEREUM_WALLET_options, $ETHEREUM_WALLET_plugin_dir, $ETHEREUM_WALLET_plugin_url_path, 
	$ETHEREUM_WALLET_services, $ETHEREUM_WALLET_amp_icons_css;

$ETHEREUM_WALLET_plugin_basename = plugin_basename( dirname( __FILE__ ) );
$ETHEREUM_WALLET_plugin_dir = untrailingslashit( plugin_dir_path( __FILE__ ) );
$ETHEREUM_WALLET_plugin_url_path = untrailingslashit( plugin_dir_url( __FILE__ ) );

// HTTPS?
$ETHEREUM_WALLET_plugin_url_path = is_ssl() ? str_replace( 'http:', 'https:', $ETHEREUM_WALLET_plugin_url_path ) : $ETHEREUM_WALLET_plugin_url_path;
// Set plugin options
$ETHEREUM_WALLET_options = get_option( 'ethereum-wallet_options', array() );

/**
 * The ERC20 smart contract ABI
 * 
 * @var string The ERC20 smart contract ABI
 * @see http://www.webtoolkitonline.com/json-minifier.html
 */
$ETHEREUM_WALLET_erc20ContractABI = '[{"constant":true,"inputs":[],"name":"totalSupply","outputs":[{"name":"supply","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"name","outputs":[{"name":"","type":"string"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"symbol","outputs":[{"name":"","type":"string"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"decimals","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{"name":"_owner","type":"address"}],"name":"balanceOf","outputs":[{"name":"balance","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{"name":"_owner","type":"address"},{"name":"_spender","type":"address"}],"name":"allowance","outputs":[{"name":"remaining","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},{"indexed":true,"name":"_spender","type":"address"},{"indexed":false,"name":"_value","type":"uint256"}],"name":"Approval","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_from","type":"address"},{"indexed":true,"name":"_to","type":"address"},{"indexed":false,"name":"_value","type":"uint256"}],"name":"Transfer","type":"event"},{"constant":false,"inputs":[{"name":"_spender","type":"address"},{"name":"_value","type":"uint256"}],"name":"approve","outputs":[{"name":"success","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"name":"_to","type":"address"},{"name":"_value","type":"uint256"}],"name":"transfer","outputs":[{"name":"success","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"name":"_from","type":"address"},{"name":"_to","type":"address"},{"name":"_value","type":"uint256"}],"name":"transferFrom","outputs":[{"name":"success","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"}]';

function ETHEREUM_WALLET_deactivate() {
    if ( ! current_user_can( 'activate_plugins' ) ) {
        return;
    }
    deactivate_plugins( plugin_basename( __FILE__ ) );
}

if ( version_compare( phpversion(), '7.0', '<' ) ) {
	add_action( 'admin_init', 'ETHEREUM_WALLET_deactivate' );
	add_action( 'admin_notices', 'ETHEREUM_WALLET_admin_notice' );
	function ETHEREUM_WALLET_admin_notice() {
		if ( ! current_user_can( 'activate_plugins' ) ) {
			return;
		}
		echo '<div class="error"><p><strong>WordPress Ethereum Wallet</strong> requires PHP version 7.0 or above.</p></div>';
		if ( isset( $_GET['activate'] ) ) {
			unset( $_GET['activate'] );
		}
	}
} else if (!function_exists('gmp_init') ) {
	add_action( 'admin_init', 'ETHEREUM_WALLET_deactivate' );
	add_action( 'admin_notices', 'ETHEREUM_WALLET_admin_notice_gmp' );
	function ETHEREUM_WALLET_admin_notice_gmp() {
		if ( ! current_user_can( 'activate_plugins' ) ) {
			return;
		}
		echo '<div class="error"><p><strong>WordPress Ethereum Wallet</strong> requires <a href="http://php.net/manual/en/book.gmp.php">GMP</a> module to be installed.</p></div>';
		if ( isset( $_GET['activate'] ) ) {
			unset( $_GET['activate'] );
		}
	}
} else if (!function_exists('mb_strtolower') ) {
	add_action( 'admin_init', 'ETHEREUM_WALLET_deactivate' );
	add_action( 'admin_notices', 'ETHEREUM_WALLET_admin_notice_mbstring' );
	function ETHEREUM_WALLET_admin_notice_mbstring() {
		if ( ! current_user_can( 'activate_plugins' ) ) {
			return;
		}
		echo '<div class="error"><p><strong>WordPress Ethereum Wallet</strong> requires <a href="http://php.net/manual/en/book.mbstring.php">Multibyte String (mbstring)</a> module to be installed.</p></div>';
		if ( isset( $_GET['activate'] ) ) {
			unset( $_GET['activate'] );
		}
	}
} else {

require $ETHEREUM_WALLET_plugin_dir . '/vendor/autoload.php';
require $ETHEREUM_WALLET_plugin_dir . '/currencyconvertor.php';
require_once $ETHEREUM_WALLET_plugin_dir . '/vendor/prospress/action-scheduler/action-scheduler.php';

function ETHEREUM_WALLET_init() {
	global $ETHEREUM_WALLET_plugin_dir,
		$ETHEREUM_WALLET_plugin_basename, 
		$ETHEREUM_WALLET_options;
	
	// Load the textdomain for translations
	load_plugin_textdomain( 'ethereum-wallet', false, $ETHEREUM_WALLET_plugin_basename . '/languages/' );
}
add_filter( 'init', 'ETHEREUM_WALLET_init' );

// Takes a hex (string) address as input
function ETHEREUM_WALLET_checksum_encode($addr_str) {
    $out = array();
    $addr = str_replace('0x', '', strtolower($addr_str));
    $addr_array = str_split($addr);
    $hash_addr = \kornrunner\Keccak::hash($addr, 256);
    $hash_addr_array = str_split($hash_addr);
    for ($idx = 0; $idx < count($addr_array); $idx++) {
        $ch = $addr_array[$idx];
        if ((int) hexdec($hash_addr_array[$idx]) >= 8) {
            $out[] = strtoupper($ch) . '';
        } else {
            $out[] = $ch . '';
        }
    }
    return '0x' . implode('', $out);
}

// create Ethereum wallet on user register
// see https://wp-kama.ru/hook/user_register
function ETHEREUM_WALLET_address_from_key($privateKeyHex) {
    $privateKey = \BitWasp\Bitcoin\Key\PrivateKeyFactory::fromHex($privateKeyHex);

    $pubKeyHex = $privateKey->getPublicKey()->getHex();
    $hash = \kornrunner\Keccak::hash(substr(hex2bin($pubKeyHex), 1), 256);
    $ethAddress = '0x' . substr($hash, 24);
    $ethAddressChkSum = ETHEREUM_WALLET_checksum_encode($ethAddress);
    
    return $ethAddressChkSum;
}

function ETHEREUM_WALLET_create_account() {
    $random = new \BitWasp\Bitcoin\Crypto\Random\Random();
    $privateKeyBuffer = $random->bytes(32);
    $privateKeyHex = $privateKeyBuffer->getHex();
    $ethAddressChkSum = ETHEREUM_WALLET_address_from_key($privateKeyHex);
    
    return [$ethAddressChkSum, $privateKeyHex];
}

// @see https://www.electrictoolbox.com/mysql-order-string-as-int/
// @see https://dev.mysql.com/doc/refman/8.0/en/fixed-point-types.html
// @see https://codex.wordpress.org/Database_Description#Table:_wp_usermeta
// @see https://codex.wordpress.org/Class_Reference/wpdb
// @see https://codex.wordpress.org/Class_Reference/wpdb#Protect_Queries_Against_SQL_Injection_Attacks
// @see https://codex.wordpress.org/Data_Validation#Database
// 
// $wpdb->esc_like( $text )
//Sanitizes $text for use in a LIKE expression of a SQL query. Will still need to be SQL escaped (with one of the above functions)
//SELECT * FROM wp_usermeta
//WHERE meta_key = 'user_ethereum_wallet_ether_balance'
//ORDER BY CAST(meta_value as DECIMAL(65)) DESC
//$fivesdrafts = $wpdb->get_results( 
//	"
//	SELECT * 
//	FROM $wpdb->posts
//	WHERE post_status = 'draft' 
//		AND post_author = 5
//	"
//);
//$results = $wpdb->get_results
//$wpdb->query( $wpdb->prepare( 
//	"
//		INSERT INTO $wpdb->postmeta
//		( post_id, meta_key, meta_value )
//		VALUES ( %d, %s, %s )
//	", 
//        10, 
//	$metakey, 
//	$metavalue 
//) );
//if ( $fivesdrafts )
//{
//foreach ( $fivesdrafts as $fivesdraft ) 
//{
//	echo $fivesdraft->post_title;
//}
//}
//else
//{
//	<h2>Not Found</h2>
//}
//
// @see https://wordpress.stackexchange.com/a/246206/137915
//$args = array(
//  'post_type' => 'services',
//  'orderby'   => 'order_clause',
//  'meta_query' => array(
//       'order_clause' => array(
//            'key' => 'order_in_archive',
//            'value' => 'some_value',
//            'type' => 'NUMERIC' // unless the field is not a number
//)));

add_action( 'rest_api_init', function () {
    register_rest_route( 'ethereumwallet/v1', '/accounts', array(
        array (
            'methods' => ['GET', 'POST'],
            'callback' => 'ETHEREUM_WALLET_get_accounts_endpoint',
            'permission_callback' => function (WP_REST_Request $request) {
                return true;
//                return current_user_can( 'edit_others_posts' );
            },
            'args' => array(
                'page' => array(
                    'description'=> esc_html__( 'Pagination argument', 'ethereum-wallet' ),
                    'type' => 'int',
                    'default' => 1,
                    'required' => false,
                    'validate_callback' => function($param, $request, $key) {
                        return is_numeric( $param );
                    },
                    'sanitize_callback' => function($param, $request, $key) {
                        return intval( $param );
                    }
                ),
                'per_page' => array(
                    'description'=> esc_html__( 'Pagination argument', 'ethereum-wallet' ),
                    'type' => 'int',
                    'default' => 20,
                    'required' => false,
                    'validate_callback' => function($param, $request, $key) {
                        return is_numeric( $param );
                    },
                    'sanitize_callback' => function($param, $request, $key) {
                        return intval( $param );
                    }
                ),
                'draw' => array(
                    'description'=> esc_html__( 'Datatables required parameter', 'ethereum-wallet' ),
                    'type' => 'int',
                    'default' => 1,
                    'required' => false,
                    'validate_callback' => function($param, $request, $key) {
                        return is_numeric( $param );
                    },
                    'sanitize_callback' => function($param, $request, $key) {
                        return intval( $param );
                    }
                ),
                's' => array(
                    'description'=> esc_html__( 'Search string', 'ethereum-wallet' ),
                    'type' => 'string',
                    'default' => "",
                    'required' => false,
                    'sanitize_callback' => function($param, $request, $key) {
                        return esc_attr( $param );
                    }
                ),
            ),
        ),
        'schema' => 'ETHEREUM_WALLET_get_accounts_schema',
    ) );
    register_rest_route( 'ethereumwallet/v1', '/web3_signTransaction', array(
        array (
            'methods' => ['POST'],
            'callback' => 'ETHEREUM_WALLET_web3_signTransaction_endpoint',
            'permission_callback' => function (WP_REST_Request $request) {
                return true;
//                return current_user_can( 'edit_others_posts' );
            },
            'args' => array(
                'from' => array(
                    'description'=> esc_html__( 'The From ethereum address', 'ethereum-wallet' ),
                    'type' => 'string',
                    'default' => "",
                    'required' => true,
                    'sanitize_callback' => function($param, $request, $key) {
                        return esc_attr( $param );
                    }
                ),
                'to' => array(
                    'description'=> esc_html__( 'The To ethereum address', 'ethereum-wallet' ),
                    'type' => 'string',
                    'default' => "",
                    'required' => true,
                    'sanitize_callback' => function($param, $request, $key) {
                        return esc_attr( $param );
                    }
                ),
                'value' => array(
                    'description'=> esc_html__( 'The Value field in hex format', 'ethereum-wallet' ),
                    'type' => 'string',
                    'default' => "",
                    'required' => true,
                    'sanitize_callback' => function($param, $request, $key) {
                        return esc_attr( $param );
                    }
                ),
                'gas' => array(
                    'description'=> esc_html__( 'The Gas Limit value in hex format', 'ethereum-wallet' ),
                    'type' => 'string',
                    'default' => "",
                    'required' => true,
                    'sanitize_callback' => function($param, $request, $key) {
                        return esc_attr( $param );
                    }
                ),
                'gasPrice' => array(
                    'description'=> esc_html__( 'The Gas Price value in hex format', 'ethereum-wallet' ),
                    'type' => 'string',
                    'default' => "",
                    'required' => true,
                    'sanitize_callback' => function($param, $request, $key) {
                        return esc_attr( $param );
                    }
                ),
                'data' => array(
                    'description'=> esc_html__( 'The Data value in hex format', 'ethereum-wallet' ),
                    'type' => 'string',
                    'default' => "",
                    'required' => true,
                    'sanitize_callback' => function($param, $request, $key) {
                        $res = esc_attr( $param );
                        if ("0x" == $res) {
                            $res = "";
                        }
                        return $res;
                    }
                ),
                'nonce' => array(
                    'description'=> esc_html__( 'The Nonce value in hex format', 'ethereum-wallet' ),
                    'type' => 'string',
                    'default' => "",
                    'required' => true,
                    'sanitize_callback' => function($param, $request, $key) {
                        return esc_attr( $param );
                    }
                ),
            ),
        ),
        'schema' => 'ETHEREUM_WALLET_web3_signTransaction_schema',
    ) );
    register_rest_route( 'ethereumwallet/v1', '/web3_signPersonalMessage', array(
        array (
            'methods' => ['POST'],
            'callback' => 'ETHEREUM_WALLET_web3_signPersonalMessage_endpoint',
            'permission_callback' => function (WP_REST_Request $request) {
                return true;
//                return current_user_can( 'edit_others_posts' );
            },
//            'args' => array(
//            ),
        ),
        'schema' => 'ETHEREUM_WALLET_web3_signPersonalMessage_schema',
    ) );
} );


function ETHEREUM_WALLET_get_accounts_schema() {
    $schema = array(
        // This tells the spec of JSON Schema we are using which is draft 4.
        '$schema'              => 'http://json-schema.org/draft-04/schema#',
        // The title property marks the identity of the resource.
        'title'                => 'EthereumAccount',
        'type'                 => 'object',
        // In JSON Schema you can specify object properties in the properties attribute.
        'properties'           => array(
            'id' => array(
                'description'  => esc_html__( 'Unique identifier for the object.', 'ethereum-wallet' ),
                'type'         => 'integer',
                'context'      => array( 'view' ),
                'readonly'     => true,
            ),
            'rank' => array(
                'description'  => esc_html__( 'The rank value.', 'ethereum-wallet' ),
                'type'         => 'integer',
            ),
            'tx_count' => array(
                'description'  => esc_html__( 'The user\'s Ethereum wallet tx count.', 'ethereum-wallet' ),
                'type'         => 'integer',
            ),
            'username' => array(
                'description'  => esc_html__( 'The user\'s login.', 'ethereum-wallet' ),
                'type'         => 'string',
            ),
            'avatarUrl' => array(
                'description'  => esc_html__( 'The user\'s avatar URL.', 'ethereum-wallet' ),
                'type'         => 'string',
            ),
            'ether_address' => array(
                'description'  => esc_html__( 'The user\'s Ethereum wallet address.', 'ethereum-wallet' ),
                'type'         => 'string',
            ),
            'ether_balance' => array(
                'description'  => esc_html__( 'The user\'s Ethereum wallet balance.', 'ethereum-wallet' ),
                'type'         => 'string',
            ),
            'fiat_balance' => array(
                'description'  => esc_html__( 'The user\'s Ethereum wallet fiat balance.', 'ethereum-wallet' ),
                'type'         => 'string',
            ),
        ),
    );
 
    return $schema;
}

function ETHEREUM_WALLET_get_accounts_endpoint( WP_REST_Request $request ) {
    global $wpdb;
    
    $blockchainNetwork = ETHEREUM_WALLET_getBlockchainNetwork();
    // Or via the helper method:
    $page = intval($request->get_param( 'page' ));
    if ($page < 1) {
        $page = 1;
    }
    $per_page = intval($request->get_param( 'per_page' ));
    $draw = intval($request->get_param( 'draw' ));
    
    $query_total = $wpdb->prepare(
        "
            SELECT COUNT(*) as num_rows
            FROM $wpdb->users AS u
            LEFT JOIN $wpdb->usermeta AS m ON u.ID=m.user_id
            WHERE m.meta_key = %s
        ", 
        'user_ethereum_wallet_ether_balance_' . $blockchainNetwork
    );
    $results_total = $wpdb->get_results( $query_total );
    $total = $results_total[0]->num_rows;
    $total_filtered = $total;

    $query = $wpdb->prepare(
        "
            SELECT 
                u.ID AS user_id,
                u.user_login AS username,
                m.meta_value AS ether_balance
            FROM $wpdb->users AS u
            LEFT JOIN $wpdb->usermeta AS m ON u.ID=m.user_id
            WHERE m.meta_key = %s
            ORDER BY CAST(m.meta_value as DECIMAL(65)) DESC
            LIMIT %d
            OFFSET %d
        ", 
        'user_ethereum_wallet_ether_balance_' . $blockchainNetwork, 
        $per_page, 
        ($page - 1) * $per_page
    );
    $search = $request->get_param( 's' );
    if (!empty($search)) {
        $searchLike = '%' . $wpdb->esc_like( $search ) . '%';
        $query_total_filtered = $wpdb->prepare(
            "
                SELECT COUNT(*) as num_rows
                FROM $wpdb->users AS u
                LEFT JOIN $wpdb->usermeta AS m ON u.ID=m.user_id
                WHERE 
                    m.meta_key = %s
                    AND u.user_login LIKE %s
            ", 
            'user_ethereum_wallet_ether_balance_' . $blockchainNetwork,
            $searchLike
        );
        $results_total_filtered = $wpdb->get_results( $query_total_filtered );
        $total_filtered = $results_total_filtered[0]->num_rows;

        $query = $wpdb->prepare(
            "
                SELECT 
                    u.ID AS user_id,
                    u.user_login AS username,
                    m.meta_value AS ether_balance
                FROM $wpdb->users AS u
                LEFT JOIN $wpdb->usermeta AS m ON u.ID=m.user_id
                WHERE 
                    m.meta_key = %s
                    AND u.user_login LIKE %s
                ORDER BY CAST(m.meta_value as DECIMAL(65)) DESC
                LIMIT %d
                OFFSET %d
            ", 
            'user_ethereum_wallet_ether_balance_' . $blockchainNetwork,
            $searchLike,
            $per_page,
            ($page - 1) * $per_page
        );
    }

    $can_list_users = current_user_can('list_users');
    
    $results = $wpdb->get_results( $query );
    if (!$results) {
        return [
            "draw" => $draw,
            "recordsTotal" => intval($total),
            "recordsFiltered" => intval($total_filtered),
            "data" => []
        ];
    }
    
    $user_ids = [];
    foreach ($results as $r) {
        $user_ids[] = $r->user_id;
    }
    $s_user_ids = implode(',', $user_ids);
    
    $meta_results = $wpdb->get_results( $wpdb->prepare( 
        "
            SELECT 
                user_id AS user_id,
                meta_key AS meta_key,
                meta_value AS meta_value
            FROM $wpdb->usermeta
            WHERE 
                user_id IN ($s_user_ids)
                AND meta_key IN (%s, %s)
        ", 
        'user_ethereum_wallet_address',
        'user_ethereum_wallet_tx_count_' . $blockchainNetwork
    ) );
    $meta = [];
    foreach ($meta_results as $r) {
        if (!isset($meta[$r->user_id])) {
            $meta[$r->user_id] = [];
        }
        $meta[$r->user_id][$r->meta_key] = $r->meta_value;
    }
    
    $convertor = null;
    if (function_exists("get_woocommerce_currency")) {
        $currency = get_woocommerce_currency();
        if ( $currency != 'MYC' && 
            !(function_exists('mycred_point_type_exists') && mycred_point_type_exists( $currency )) 
        ) {
            $cryptocompareApiKey = '';
            if (isset($ETHEREUM_WALLET_options['cryptocompare_api_key'])) {
                $cryptocompareApiKey = esc_attr($ETHEREUM_WALLET_options['cryptocompare_api_key']);
            }
            $convertor = new \ETHEREUM_WALLET\CurrencyConvertor('ETH', $currency, $cryptocompareApiKey);
        }
    }
    $powDecimals = new phpseclib\Math\BigInteger(pow(10, 18));
    // Rank | 
    // User profile image (VIA BUDDYPRESS) | 
    // Username (VIA BUDDYPRESS) | 
    // Wallet ETH Address | 
    // ETH FIAT Balance in Wallet (Example: $357.89) | 
    // TX Count
    $resp = [];
    foreach ($results as $idx => $r) {
        $rank = $idx + 1 + ($page - 1) * $per_page;
        
        $accountAddress = isset($meta[$r->user_id]['user_ethereum_wallet_address']) ?
            $meta[$r->user_id]['user_ethereum_wallet_address'] : '';
        $txCount = isset($meta[$r->user_id]['user_ethereum_wallet_tx_count_' . $blockchainNetwork]) ?
            $meta[$r->user_id]['user_ethereum_wallet_tx_count_' . $blockchainNetwork] : '';

        if (function_exists('bp_core_fetch_avatar')) {
            // @see https://stackoverflow.com/a/7370709/4256005
            // @see http://hookr.io/functions/bp_core_fetch_avatar/
            $avatarUrl = bp_core_fetch_avatar ( array( 'item_id' => $r->user_id, 'html' => false ) );
        } else {
            $avatarUrl = get_avatar_url($r->user_id);
        }

        $res = [
            'rank' => $rank
            , 'ether_balance' => $r->ether_balance
            , 'ether_address' => $accountAddress
            , 'tx_count' => $txCount
        ];
        if ($can_list_users) {
            $res['id'] = $r->user_id;
            $res['username'] = $r->username;
            $res['avatarUrl'] = $avatarUrl;
        }
        $fiatBalance = 0;
        if ("0" !== $r->ether_balance && !is_null($convertor)) {
            $balance = new phpseclib\Math\BigInteger($r->ether_balance);
            list($q, $r) = $balance->divide($powDecimals);
            $sR = $r->toString();
            $strBalanceDecimals = sprintf('%018s', $sR);
            $strBalance = $q->toString() . '.' . $strBalanceDecimals;
            $strBalance = rtrim($strBalance, '0');
            $strBalance = rtrim($strBalance, '.');

            $fiatBalance = $convertor->convert(floatval($strBalance));
            $fiatBalance = 0.01 * floor(100 * $fiatBalance);
        }
        if (function_exists("wc_price")) {
            $fiatBalance = wc_price($fiatBalance);
        }
        $res['fiat_balance'] = $fiatBalance;
        $resp[] = $res;
    }
    return [
        "draw" => $draw,
        "recordsTotal" => intval($total),
        "recordsFiltered" => intval($total_filtered),
        "data" => $resp
    ];
}


function ETHEREUM_WALLET_web3_signTransaction_schema() {
    $schema = array(
        // This tells the spec of JSON Schema we are using which is draft 4.
        '$schema'              => 'http://json-schema.org/draft-04/schema#',
        // The title property marks the identity of the resource.
        'title'                => 'TxRawSigned',
        'type'                 => 'string'
    );
 
    return $schema;
}

function ETHEREUM_WALLET_web3_signTransaction_endpoint( WP_REST_Request $request ) {
//    $blockchainNetwork = ETHEREUM_WALLET_getBlockchainNetwork();
    $params = $request->get_params();
    ETHEREUM_WALLET_log("ETHEREUM_WALLET_web3_signTransaction_endpoint: " . print_r($params, true));
    unset($params['_wpnonce']);
    
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_web3_signTransaction_endpoint: no user is logged in");
        return;
    }
    $from = get_user_meta( $user_id, 'user_ethereum_wallet_address', true);
    if (empty($from)) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_web3_signTransaction_endpoint: empty user_ethereum_wallet_address address");
        return '';
    }
    if ($from != $params['from']) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_web3_signTransaction_endpoint: from address is invalid");
        return '';
    }
    ob_start();
    $txRawSigned = ETHEREUM_WALLET_sign_transaction(
        $params['to'], 
        $params['value'], 
        $params['data'], 
        $params['gas'], 
        $params['gasPrice'], 
        $params['nonce']
    );
    $errors = ob_get_contents();
    if (!empty($errors)) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_web3_signTransaction_endpoint PHP Warnings: " . $errors);
    }
    ob_end_clean();
    if (false === $txRawSigned) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_web3_signTransaction_endpoint: empty ETHEREUM_WALLET_send_transaction result");
        return '';
    }
    ETHEREUM_WALLET_log("ETHEREUM_WALLET_web3_signTransaction_endpoint: signed raw tx: " . $txRawSigned);
    return $txRawSigned;
}

function ETHEREUM_WALLET_web3_signPersonalMessage_schema() {
    $schema = array(
        // This tells the spec of JSON Schema we are using which is draft 4.
        '$schema'              => 'http://json-schema.org/draft-04/schema#',
        // The title property marks the identity of the resource.
        'title'                => 'TxRawMsgSigned',
        'type'                 => 'string'
    );
 
    return $schema;
}

function ETHEREUM_WALLET_web3_signPersonalMessage_endpoint( WP_REST_Request $request ) {
//    $blockchainNetwork = ETHEREUM_WALLET_getBlockchainNetwork();
    $params = $request->get_body_params();
    ETHEREUM_WALLET_log("ETHEREUM_WALLET_web3_signPersonalMessage_endpoint: ", print_r($params, true));
    unset($params['_wpnonce']);
    
    $txRawSigned = '';
    return $txRawSigned;
}

function ETHEREUM_WALLET_accounts_table_shortcode( $attrs ) {
    global $ETHEREUM_WALLET_options;

//    $user_id = get_current_user_id();
//    if ( $user_id <= 0 ) {
//        ETHEREUM_WALLET_log("No user is logged in");
//        return;
//    }
    $can_list_users = current_user_can('list_users');
	$js = '';
    // Rank | 
    // User profile image (VIA BUDDYPRESS) | 
    // Username (VIA BUDDYPRESS) | 
    // Wallet ETH Address | 
    // ETH FIAT Balance in Wallet (Example: $357.89) | 
    // TX Count
    if ($can_list_users) {
        
	$ret = 
'<div class="twbs"><div class="container-fluid ethereum-wallet-accounts-table-shortcode">
    <div class="row ethereum-wallet-accounts-table-wrapper">
        <div class="col-12">
            <div class="table-responsive">
                <table 
                    class="table table-striped table-condensed ethereum-wallet-accounts-table"
                    data-wpnonce="'.wp_create_nonce( 'wp_rest' ).'">
                    <thead>
                        <tr>
                            <th>'.__('Rank', 'ethereum-wallet').'</th>
                            <th>'.__('Avatar', 'ethereum-wallet').'</th>
                            <th>'.__('Username', 'ethereum-wallet').'</th>
                            <th>'.__('Wallet', 'ethereum-wallet').'</th>
                            <th>'.__('Balance', 'ethereum-wallet').'</th>
                            <th>'.__('Tx Count', 'ethereum-wallet').'</th>
                        </tr>
                    </thead>
                    <tbody>
                    </tbody>
                </table>
            </div>
        </div>
    </div>
</div></div>';
    } else {
	$ret = 
'<div class="twbs"><div class="container-fluid ethereum-wallet-accounts-table-shortcode">
    <div class="row ethereum-wallet-accounts-table-wrapper">
        <div class="col-12">
            <div class="table-responsive">
                <table 
                    class="table table-striped table-condensed ethereum-wallet-accounts-table"
                    data-wpnonce="'.wp_create_nonce( 'wp_rest' ).'">
                    <thead>
                        <tr>
                            <th>'.__('Rank', 'ethereum-wallet').'</th>
                            <th>'.__('Wallet', 'ethereum-wallet').'</th>
                            <th>'.__('Balance', 'ethereum-wallet').'</th>
                            <th>'.__('Tx Count', 'ethereum-wallet').'</th>
                        </tr>
                    </thead>
                    <tbody>
                    </tbody>
                </table>
            </div>
        </div>
    </div>
</div></div>';
    }

    ETHEREUM_WALLET_enqueue_scripts_();
//    wp_enqueue_script( 'wp-api' );

    return $js . str_replace("\n", " ", str_replace("\r", " ", str_replace("\t", " ", $js . $ret)));
}

add_shortcode( 'ethereum-wallet-accounts-table', 'ETHEREUM_WALLET_accounts_table_shortcode' );



add_action("ETHEREUM_WALLET_adjust_balance", "ETHEREUM_WALLET_adjust_balance", 0, 1);

function ETHEREUM_WALLET_get_txcount($accountAddress) {
    $providerUrl = ETHEREUM_WALLET_getWeb3Endpoint();
    $requestManager = new \Web3\RequestManagers\HttpRequestManager($providerUrl, 10/* seconds */);
    $web3 = new \Web3\Web3(new \Web3\Providers\HttpProvider($requestManager));
    $eth = $web3->eth;
    $txCount = 0;
    $eth->getTransactionCount($accountAddress, function ($err, $transactionCount) use(&$txCount) {
        if ($err !== null) {
            ETHEREUM_WALLET_log("Failed to getTransactionCount: " . $err);
            $txCount = null;
            return;
        }
        $txCount = intval($transactionCount->toString());
        if ($txCount > 0) {
            $txCount = $txCount - 1;
        }
    });
    return $txCount;
}

function ETHEREUM_WALLET_adjust_balance($user_id) {
    $accountAddress = get_user_meta( $user_id, 'user_ethereum_wallet_address', true);
    if (empty($accountAddress)) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_adjust_balance: user_ethereum_wallet_address is empty for $user_id");
        return;
    }
    $providerUrl = ETHEREUM_WALLET_getWeb3Endpoint();
    $balance = ETHEREUM_WALLET_getBalanceEth($providerUrl, $accountAddress);
    if (!is_null($balance)) {
        $strBalance = $balance->toString();
        $blockchainNetwork = ETHEREUM_WALLET_getBlockchainNetwork();
        $oldBalance = get_user_meta( $user_id, 'user_ethereum_wallet_ether_balance_' . $blockchainNetwork, true);
        if ($oldBalance != $strBalance) {
            update_user_meta( $user_id, 'user_ethereum_wallet_ether_balance_' . $blockchainNetwork, $strBalance);
            ETHEREUM_WALLET_log("ETHEREUM_WALLET_adjust_balance: new balance $strBalance (old: $oldBalance) for $user_id");
        }
    } else {
        ETHEREUM_WALLET_log("Failed to retrieve Ether balance for $user_id");
        $offset = 30;
        ETHEREUM_WALLET_enqueue_adjust_balance_task($user_id, $offset);
    }
    
    $txCount = ETHEREUM_WALLET_get_txcount($accountAddress);
    if (!is_null($txCount)) {
        $blockchainNetwork = ETHEREUM_WALLET_getBlockchainNetwork();
        $oldTxCount = get_user_meta( $user_id, 'user_ethereum_wallet_tx_count_' . $blockchainNetwork, true);
        if ($oldTxCount != $txCount) {
            update_user_meta( $user_id, 'user_ethereum_wallet_tx_count_' . $blockchainNetwork, $txCount);
            ETHEREUM_WALLET_log("ETHEREUM_WALLET_adjust_balance: new tx count $txCount (old: $oldTxCount) for $user_id");
        }
    } else {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_adjust_balance: failed to obtain tx count for $user_id");
        if (!is_null($balance)) {
            $offset = 30;
            ETHEREUM_WALLET_enqueue_adjust_balance_task($user_id, $offset);
        }
    }
}

function ETHEREUM_WALLET_enqueue_adjust_balance_task($user_id, $offset = 0) {
    global $ETHEREUM_WALLET_options;
    
    $timestamp = time() + $offset;
    $hook = "ETHEREUM_WALLET_adjust_balance";
    $args = array($user_id);
    $group = $user_id . "-ethereum-wallet-queue";
    $next_action = as_next_scheduled_action( $hook, $args, $group );
    if (false === $next_action) {
        $task_id = as_schedule_single_action( $timestamp, $hook, $args, $group );
//        ETHEREUM_WALLET_log("Task adjust_balance with id $task_id for order $user_id scheduled for group: $group");
    } else {
//        ETHEREUM_WALLET_log("Task adjust_balance already scheduled for group $group at $next_action time");
    }
}


function ETHEREUM_WALLET_user_ping( ) {
    global $wp;
    
    if (isset($_REQUEST['_wpnonce'])) {
        $nonce_value = $_REQUEST['_wpnonce'];
        if ( ! wp_verify_nonce( $nonce_value ) ) {
//            ETHEREUM_WALLET_log("ETHEREUM_WALLET_user_ping: bad nonce detected: " . $nonce_value);
            return;
        }
    }
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    $offset = 30;
    ETHEREUM_WALLET_enqueue_adjust_balance_task($user_id, $offset);
}
add_action('wp_loaded', 'ETHEREUM_WALLET_user_ping', 20);
//add_action('wp_login', 'ETHEREUM_WALLET_user_login', 10, 2);

// create Ethereum wallet on user register
// see https://wp-kama.ru/hook/user_register
add_action( 'user_register', 'ETHEREUM_WALLET_user_registration' );
function ETHEREUM_WALLET_user_registration( $user_id ) {
    $accountsJSON = get_user_meta( $user_id, 'user_ethereum_wallet_accounts', true);
    if (!empty($accountsJSON)) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_user_registration: account already exists");
        return;
    }
    list($ethAddressChkSum, $privateKeyHex) = ETHEREUM_WALLET_create_account();
    $accounts = [[
        "name" => __( 'Default account', 'ethereum-wallet' ),
        "address" => $ethAddressChkSum,
        "key" => $privateKeyHex,
    ]];
    update_user_meta( $user_id, 'user_ethereum_wallet_accounts', json_encode($accounts));

    // set default account
    update_user_meta( $user_id, 'user_ethereum_wallet_address', $ethAddressChkSum);
    update_user_meta( $user_id, 'user_ethereum_wallet_key', $privateKeyHex);
}

function ETHEREUM_WALLET_calc_display_value( $value ) {
    if ($value < 1) {
        return [0.01 * floor(100 * $value), __( 'ETH', 'ethereum-wallet' )];
    }
    if ($value < 1000) {
        return [0.1 * floor(10 * $value), __( 'ETH', 'ethereum-wallet' )];
    }
    if ($value < 1000000) {
        return [0.1 * floor(10 * 0.001 * $value), __( 'K', 'ethereum-wallet' )];
    }
    return [0.1 * floor(10 * 0.000001 * $value), __( 'M', 'ethereum-wallet' )];
}

class ETHEREUM_WALLET_Logger {

    /**
     * Add a log entry.
     *
     * This is not the preferred method for adding log messages. Please use log() or any one of
     * the level methods (debug(), info(), etc.). This method may be deprecated in the future.
     *
     * @param string $handle
     * @param string $message
     * @param string $level
     * 
     * @see https://docs.woocommerce.com/wc-apidocs/source-class-WC_Logger.html#105
     *
     * @return bool
     */
    public function add( $handle, $message, $level = 'unused' ) {
        error_log($handle . ': ' . $message);
        return true;
    }
}

function ETHEREUM_WALLET_log($error) {
    static $logger = false;
    // Create a logger instance if we don't already have one.
    if ( false === $logger ) {
        /**
         * Check if WooCommerce is active
         * https://wordpress.stackexchange.com/a/193908/137915
         **/
        if ( 
          in_array( 
            'woocommerce/woocommerce.php', 
            apply_filters( 'active_plugins', get_option( 'active_plugins' ) ) 
          ) 
        ) {
            $logger = new WC_Logger();
        } else {
            $logger = new ETHEREUM_WALLET_Logger();
        }
    }
    $logger->add( 'ethereum-wallet', $error );
}

function ETHEREUM_WALLET_getBalanceEth($providerUrl, $accountAddress) {
    $requestManager = new \Web3\RequestManagers\HttpRequestManager($providerUrl, 10/* seconds */);
    $web3 = new \Web3\Web3(new \Web3\Providers\HttpProvider($requestManager));
    $eth = $web3->eth;

    $ether_quantity_wei = null;
    $eth->getBalance($accountAddress, function ($err, $balance) use(&$ether_quantity_wei) {
        if ($err !== null) {
            ETHEREUM_WALLET_log("Failed to getBalance: " . $err);
            return;
        }
        $ether_quantity_wei = $balance;
    });
    return $ether_quantity_wei;
}

function ETHEREUM_WALLET_getWeb3Endpoint() {
    global $ETHEREUM_WALLET_options;
    $infuraApiKey = '';
    if (isset($ETHEREUM_WALLET_options['infuraApiKey'])) {
        $infuraApiKey = esc_attr($ETHEREUM_WALLET_options['infuraApiKey']);
    }
    $blockchainNetwork = ETHEREUM_WALLET_getBlockchainNetwork();
    $web3Endpoint = "https://" . esc_attr($blockchainNetwork) . ".infura.io/v3/" . esc_attr($infuraApiKey);
    return $web3Endpoint;
}

function ETHEREUM_WALLET_getBlockchainNetwork() {
    global $ETHEREUM_WALLET_options;
    if (!isset($ETHEREUM_WALLET_options['blockchain_network'])) {
        return 'mainnet';
    }
    $blockchainNetwork = esc_attr($ETHEREUM_WALLET_options['blockchain_network']);
    if (empty($blockchainNetwork)) {
        $blockchainNetwork = 'mainnet';
    }
    return $blockchainNetwork;
}

function ETHEREUM_WALLET_balance_shortcode( $attrs ) {
    global $ETHEREUM_WALLET_options;

    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    $accountAddress = ETHEREUM_WALLET_get_wallet_address();
    
    $attributes = shortcode_atts( array(
        'tokenaddress' => '',
        'tokendecimals' => '2',
        'tokendecimalchar' => '.',
        'tokenwooproduct' => '',
        'tokeniconpath' => '',
        'displayfiat' => '0',
    ), $attrs, 'ethereum-wallet' );

    $tokenAddress = ! empty( $attributes['tokenaddress'] ) ? esc_attr($attributes['tokenaddress']) : '';
    $tokenDecimals = intval(isset( $attributes['tokendecimals'] ) ? esc_attr($attributes['tokendecimals']) : '2');
    $tokenDecimalCharDefault = '.';
    if (function_exists("wc_get_price_decimal_separator")) {
        $tokenDecimalCharDefault = wc_get_price_decimal_separator();
    }
    $tokenDecimalChar = ! empty( $attributes['tokendecimalchar'] ) ? esc_attr($attributes['tokendecimalchar']) : $tokenDecimalCharDefault;
    $product_id = ! empty( $attributes['tokenwooproduct'] ) ? intval(esc_attr($attributes['tokenwooproduct'])) : '';
    $tokenIconPath = ! empty( $attributes['tokeniconpath'] ) ? esc_attr($attributes['tokeniconpath']) : '';
    $displayFiat = ! empty( $attributes['displayfiat'] ) ? boolval(esc_attr($attributes['displayfiat'])) : false;

    $providerUrl = ETHEREUM_WALLET_getWeb3Endpoint();
    
    /**
     * @param phpseclib\Math\BigInteger $balance The Ether or Token balance in wei.
     */
    $balance = new phpseclib\Math\BigInteger(0);
    $strBalance = '0';
    $strCurrencyName = __('Ethereum', 'ethereum-wallet');
    $strCurrencySymbol = "ETH";
    if (!empty($accountAddress)) {
        // ETH
        $balance = ETHEREUM_WALLET_getBalanceEth($providerUrl, $accountAddress);
        if (!is_null($balance)) {
            $powDecimals = new phpseclib\Math\BigInteger(pow(10, 18));
            list($q, $r) = $balance->divide($powDecimals);
            if (0 == $tokenDecimals) {
                if ($r->multiply(new phpseclib\Math\BigInteger(2))->compare($powDecimals) < 0) {
                    $strBalance = $q->toString();
                } else {
                    $strBalance = $q->add(new phpseclib\Math\BigInteger(1))->toString();
                }
            } else {
                $sR = $r->toString();
                $strBalanceDecimals = sprintf('%018s', $sR);
                $strBalanceDecimals2 = substr($strBalanceDecimals, 0, $tokenDecimals);
                if (str_pad("", $tokenDecimals, "0") == $strBalanceDecimals2) {
                    $strBalance = rtrim($q->toString() . $tokenDecimalChar . $strBalanceDecimals, '0');
                } else {
                    $strBalance = rtrim($q->toString() . $tokenDecimalChar . $strBalanceDecimals2, '0');
                }
                $strBalance = rtrim($strBalance, $tokenDecimalChar);
            }
        } else {
            $strBalance = __('Failed to retrieve Ether balance', 'ethereum-wallet');
        }
        $strBalance = rtrim($strBalance, '.');
    }

    $js = '';

    $fiatBalance = null;
    $exchangeRate = null;
    if ($product_id) {
        if (function_exists("wc_get_product")) {
            $product = wc_get_product($product_id);
            if ($product) {
                $exchangeRate = $product->get_regular_price();
                if (function_exists("CRYPTOCURRENCY_PRODUCT_FOR_WOOCOMMERCE_get_product_price")) {
                    $exchangeRate = CRYPTOCURRENCY_PRODUCT_FOR_WOOCOMMERCE_get_product_price($exchangeRate, $product);
                }
                $fiatBalance = 0.01 * floor(100 * floatval($strBalance) * floatval($exchangeRate));
                if (function_exists("wc_price")) {
                    $fiatBalance = wc_price($fiatBalance);
                    $exchangeRate = wc_price($exchangeRate);
                }
            }
        }
    } else if ("ETH" == $strCurrencySymbol && $displayFiat) {
        if (function_exists("get_woocommerce_currency")) {
            $currency = get_woocommerce_currency();
            if ( $currency != 'MYC' && 
                !(function_exists('mycred_point_type_exists') && mycred_point_type_exists( $currency )) 
            ) {
                $cryptocompareApiKey = '';
                if (isset($ETHEREUM_WALLET_options['cryptocompare_api_key'])) {
                    $cryptocompareApiKey = esc_attr($ETHEREUM_WALLET_options['cryptocompare_api_key']);
                }
                $convertor = new \ETHEREUM_WALLET\CurrencyConvertor('ETH', $currency, $cryptocompareApiKey);
                $exchangeRate = $convertor->get_exchange_rate();
                $fiatBalance = $convertor->convert(floatval($strBalance));
                $fiatBalance = 0.01 * floor(100 * $fiatBalance);
                if (function_exists("wc_price")) {
                    $fiatBalance = wc_price($fiatBalance);
                    $exchangeRate = wc_price($exchangeRate);
                }
            }
        }
    }
    
    if (is_null($fiatBalance)) {
        $ret = 
'<div class="twbs"><div class="container-fluid ethereum-wallet-balance-shortcode">
    <div class="row ethereum-wallet-balance-content">
        <div class="col-md-6 col-6 text-right ethereum-wallet-balance-value-wrapper">
            <div class="ethereum-wallet-balance-value">'.$strBalance.'</div>
        </div>
        <div class="col-md-6 col-6 ethereum-wallet-balance-token-name-wrapper">
            <div style="display: none" class="hidden ethereum-wallet-balance-token-address">'.$tokenAddress.'</div>
            <div style="display: none" class="hidden ethereum-wallet-balance-token-decimals">'.$tokenDecimals.'</div>
            <div style="display: none" class="hidden ethereum-wallet-balance-token-decimal-char">'.$tokenDecimalChar.'</div>
            <div class="ethereum-wallet-balance-token-name">'.$strCurrencySymbol.'</div>
        </div>
    </div>
</div></div>';
    } else if (empty($tokenIconPath)) {
        $ret = 
'<div class="twbs"><div class="container-fluid ethereum-wallet-balance-shortcode">
    <div class="row ethereum-wallet-balance-content">
        <div class="col-md-4 col-4 text-right ethereum-wallet-balance-value-wrapper">
            <div class="ethereum-wallet-balance-value">'.$strBalance.'</div>
        </div>
        <div class="col-md-4 col-4 ethereum-wallet-balance-token-name-wrapper">
            <div style="display: none" class="hidden ethereum-wallet-balance-token-address">'.$tokenAddress.'</div>
            <div style="display: none" class="hidden ethereum-wallet-balance-token-decimals">'.$tokenDecimals.'</div>
            <div style="display: none" class="hidden ethereum-wallet-balance-token-decimal-char">'.$tokenDecimalChar.'</div>
            <div class="ethereum-wallet-balance-token-name">'.$strCurrencySymbol.'</div>
        </div>
        <div class="col-md-4 col-4 ethereum-wallet-balance-fiat-value-wrapper">
            <div class="ethereum-wallet-balance-fiat-value">'.$fiatBalance.'</div>
        </div>
    </div>
</div></div>';
    } else {

        // @see https://bootsnipp.com/snippets/org5r
        $ret = 
'<div class="twbs"><div class="container-fluid ethereum-wallet-balance-shortcode"><ul class="list-group">
    <li class="list-group-item ethereum-wallet-balance-content">
        <div class="col-7 col-xs-7 col-sm-7 ethereum-wallet-balance-wrapper-left">
            <div class="ethereum-wallet-balance-token-image-wrapper">
                <img src="'.$tokenIconPath.'" class="ethereum-wallet-balance-token-image">
            </div>
            <span class="ethereum-wallet-balance-token-long-name">'.$strCurrencyName.'</span><br>
            <span class="small ethereum-wallet-balance-exchange-rate">'.
                sprintf(__('%1$s per %2$s', 'ethereum-wallet'), $exchangeRate, $strCurrencySymbol).
            '</span>
        </div>
        <div style="display: none" class="hidden ethereum-wallet-balance-token-address">'.$tokenAddress.'</div>
        <div style="display: none" class="hidden ethereum-wallet-balance-token-symbol">'.$strCurrencySymbol.'</div>
        <div style="display: none" class="hidden ethereum-wallet-balance-token-decimals">'.$tokenDecimals.'</div>
        <div style="display: none" class="hidden ethereum-wallet-balance-token-decimal-char">'.$tokenDecimalChar.'</div>
        <div class="col-5 col-xs-5 col-sm-5 text-right ethereum-wallet-balance-wrapper-right">
            <span class="ethereum-wallet-balance-fiat-value">'.$fiatBalance.'</span><br>
            <span class="ethereum-wallet-balance-value">'.sprintf(__('%1$s %2$s', 'ethereum-wallet'), $strBalance, $strCurrencySymbol).'</span>
        </div>
        <div class="clearfix"></div>
    </li>
</ul></div></div>';
        
    }

    ETHEREUM_WALLET_enqueue_scripts_();

    return $js . str_replace("\n", " ", str_replace("\r", " ", str_replace("\t", " ", $js . $ret)));
}

add_shortcode( 'ethereum-wallet-balance', 'ETHEREUM_WALLET_balance_shortcode' );

function ETHEREUM_WALLET_account_shortcode( $attrs ) {
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    static $counter = 0;
    ++$counter;
    $accountAddress = ETHEREUM_WALLET_get_wallet_address();

    $attributes = shortcode_atts( array(
        'label' => '',
        'nolabel' => '',
    ), $attrs, 'ethereum-wallet' );

    $label = ! empty( $attributes['label'] ) ? esc_attr($attributes['label']) : __('Account', 'ethereum-wallet');
    $nolabel = ! empty( $attributes['nolabel'] ) ? esc_attr($attributes['nolabel']) : '';
    
    $labelTag = '<label class="control-label" for="ethereum-wallet-account">'.$label.'</label>';
    if ('yes' == $nolabel) {
        $labelTag = '';
    }
    
    $js = '';
    $ret = 
'<div class="twbs"><div class="container-fluid ethereum-wallet-account-shortcode">
    <div class="row ethereum-wallet-account-account-wrapper ethereum-wallet-account-wrapper">
        <div class="col-12">
            <div class="form-group">
                '.$labelTag.'
                <div class="input-group" style="margin-top: 8px">
                    <input type="text"
                           value="'.$accountAddress.'" 
                           readonly 
                           id="ethereum-wallet-account' . $counter . '" 
                           class="form-control">
                    <span class="input-group-append">
                        <div class="btn-group" role="group">
                            <button class="button btn btn-default btn-left d-md-inline ethereum-wallet-qr-button" 
                                    type="button" 
                                    data-toggle="collapse" 
                                    href="#ethereum-wallet-account-qr' . $counter . '" 
                                    role="button" 
                                    aria-expanded="false" 
                                    aria-controls="ethereum-wallet-account-qr' . $counter . '" 
                                    title="' . __('QR', 'ethereum-wallet') . '">
                                <i class="fa fa-qrcode" aria-hidden="true"></i>
                            </button>
                            <button class="button btn btn-default btn-right ethereum-wallet-copy-button" type="button"
                                    data-clipboard-target="#ethereum-wallet-account' . $counter . '"
                                    data-clipboard-action="copy" 
                                    id="ethereum-wallet-account-copy-qr-button' . $counter . '"
                                    title="'. __('Copy', 'ethereum-wallet') . '">
                                <i class="fa fa-clipboard" aria-hidden="true"></i>
                            </button>
                        </div>
                    </span>
                </div>
                <div class="collapse ethereum-wallet-account-qr" 
                        id="ethereum-wallet-account-qr' . $counter . '">
                    <div class="d-md-block mx-auto col-md-4 float-none">
                        <div class="ethereum-wallet-account-canvas-qr' . $counter . '"></div>
                    </div>
                </div>
            </div>
        </div>
    </div>
</div></div>';

    ETHEREUM_WALLET_enqueue_scripts_();
    wp_enqueue_script( 'jquery.qrcode' );
    wp_enqueue_script( 'clipboard' );

    return $js . str_replace("\n", " ", str_replace("\r", " ", str_replace("\t", " ", $js . $ret)));
}

add_shortcode( 'ethereum-wallet-account', 'ETHEREUM_WALLET_account_shortcode' );

function ETHEREUM_WALLET_account_management_export_shortcode( $attrs ) {
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    static $counter = 0;
    ++$counter;
    $privateKey = get_user_meta( $user_id, 'user_ethereum_wallet_key', true);
    if (empty($privateKey)) {
        ETHEREUM_WALLET_user_registration( $user_id );
        $privateKey = get_user_meta( $user_id, 'user_ethereum_wallet_key', true);
    }

    $attributes = shortcode_atts( array(
        'label' => '',
        'nolabel' => '',
    ), $attrs, 'ethereum-wallet' );

    $label = ! empty( $attributes['label'] ) ? esc_attr($attributes['label']) : __('Private key', 'ethereum-wallet');
    $nolabel = ! empty( $attributes['nolabel'] ) ? esc_attr($attributes['nolabel']) : '';
    
    $labelTag = '<label class="control-label" for="ethereum-wallet-account-management-export">'.$label.'</label>';
    if ('yes' == $nolabel) {
        $labelTag = '';
    }
    
    $js = '';
    $ret = 
'<div class="twbs"><div class="container-fluid ethereum-wallet-account-management-export-shortcode">
    <div class="row ethereum-wallet-account-management-export-wrapper">
        <div class="col-12">
            <div class="form-group">
                '.$labelTag.'
                <div class="input-group" style="margin-top: 8px">
                    <input type="text"
                           value="'.$privateKey.'" 
                           readonly 
                           id="ethereum-wallet-account-management-export' . $counter . '" 
                           class="form-control">
                    <span class="input-group-append">
                        <div class="btn-group" role="group">
                            <button class="button btn btn-default btn-left d-md-inline ethereum-wallet-qr-button" type="button" 
                                    data-toggle="collapse" 
                                    href="#ethereum-wallet-account-management-export-qr' . $counter . '" 
                                    role="button" aria-expanded="false" 
                                    aria-controls="ethereum-wallet-account-management-export-qr' . $counter . '" 
                                    title="' . __('QR', 'ethereum-wallet') . '">
                                <i class="fa fa-qrcode" aria-hidden="true"></i>
                            </button>
                            <button class="button btn btn-default btn-right ethereum-wallet-copy-button" type="button"
                                    data-clipboard-target="#ethereum-wallet-account-management-export' . $counter . '"
                                    data-clipboard-action="copy" 
                                    title="'. __('Copy', 'ethereum-wallet') . '"
                                    id="ethereum-wallet-account-management-export-copy-qr-button' . $counter . '">
                                <i class="fa fa-clipboard" aria-hidden="true"></i>
                            </button>
                        </div>
                    </span>
                </div>
                <div class="collapse ethereum-wallet-account-management-export-qr" 
                        id="ethereum-wallet-account-management-export-qr' . $counter . '">
                    <div class="d-md-block mx-auto col-md-4 float-none">
                        <div class="ethereum-wallet-account-management-export-canvas-qr' . $counter . '"></div>
                    </div>
                </div>
            </div>
        </div>
    </div>
</div></div>';

    ETHEREUM_WALLET_enqueue_scripts_();
    wp_enqueue_script( 'jquery.qrcode' );
    wp_enqueue_script( 'clipboard' );

    return $js . str_replace("\n", " ", str_replace("\r", " ", str_replace("\t", " ", $js . $ret)));
}

add_shortcode( 'ethereum-wallet-account-management-export', 'ETHEREUM_WALLET_account_management_export_shortcode' );

function ETHEREUM_WALLET_sendform_shortcode( $attributes ) {
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    $accountAddress = ETHEREUM_WALLET_get_wallet_address();

    $ops = '';
    
    $lastTxHash = get_user_meta( $user_id, 'user_ethereum_wallet_last_txhash', true);
    $lastTxTime = get_user_meta( $user_id, 'user_ethereum_wallet_last_txtime', true);

    if (!empty($lastTxHash) && 0 != ETHEREUM_WALLET_is_tx_confirmed($lastTxHash)) {
        delete_user_meta( $user_id, 'user_ethereum_wallet_last_txhash', $lastTxHash);
        delete_user_meta( $user_id, 'user_ethereum_wallet_last_txtime', $lastTxTime);
        $lastTxHash = '';
        $lastTxTime = '';
    }
    
    if (!empty($lastTxHash)) {
        $lastTxMsg = sprintf(
            __('Last transaction %1$s%2$s%3$s%4$s%5$s is still in progress. Wait please.', 'ethereum-wallet')
            , '<a target="_blank" href="'
            , ETHEREUM_WALLET_get_txhash_path($lastTxHash)
            , '">'
            , substr($lastTxHash, 0, 8)
            , '</a>'
        );
    } else {
        $lastTxMsg = sprintf(
            __('Last transaction %1$s%2$s%3$s%4$s%5$s is still in progress. Wait please.', 'ethereum-wallet')
            , ''
            , ''
            , ''
            , ''
            , ''
        );
    }

	$js = '';
	$ret = 
'<form method="post" action="" onsubmit="return ETHEREUM_WALLET_validate_send_form()"><div class="twbs"><div class="container-fluid ethereum-wallet-sendform-shortcode">
    <div class="row ethereum-wallet-sendform-content">
        <div class="col-12">
            <div id="ethereum-wallet-tx-in-progress-alert" class="form-group hidden" hidden>
                <div class="alert alert-warning" role="alert">
                    <!--http://jsfiddle.net/0vzmmn0v/1/-->
                    <div class="fa fa-exclamation-triangle" aria-hidden="true"></div>
                    <div>' .$lastTxMsg. '</div>
                </div>
            </div>
        </div>
    </div>
    <div class="row ethereum-wallet-sendform-content">
        <div class="col-12">
            <div class="form-group">
                <label class="control-label" for="ethereum-wallet-sendform-to">'.__('To', 'ethereum-wallet').'</label>
                <div class="input-group" style="margin-top: 8px">
                    <input type="text"
                           value="" 
                           placeholder="'.__('Input the recipient ethereum address', 'ethereum-wallet').'" 
                           id="ethereum-wallet-sendform-to" 
                           name="ethereum-wallet-sendform-to" 
                           class="form-control">
                    <span class="input-group-append">
                        <div class="btn-group" role="group">
                            <button class="button btn btn-default btn-left d-md-inline ethereum-wallet-qr-scan-button" type="button" 
                                    data-toggle="collapse" 
                                    href="#ethereum-wallet-to-qr1" 
                                    role="button" aria-expanded="false" 
                                    aria-controls="ethereum-wallet-to-qr1" 
                                    title="' . __('QR', 'ethereum-wallet') . '">
                                <i class="fa fa-qrcode" aria-hidden="true"></i>
                            </button>
                        </div>
                    </span>
                </div>
                <div class="collapse" id="ethereum-wallet-to-qr1">
                    <div class="d-md-block mx-auto col-md-4 float-none">
                        <div class="d-none" id="ethereum-wallet-to-canvas-qr1-loading-message">🎥' . 
                            __('Unable to access video stream (please make sure you have a webcam enabled)', 'ethereum-wallet') . 
                        '</div>
                        <canvas id="ethereum-wallet-to-canvas-qr1" width="256px" height="256px" hidden></canvas>
                    </div>
                </div>
            </div>
        </div>
    </div>
    <div class="row ethereum-wallet-sendform-content">
        <div class="col-12">
            <div class="form-group">
                <label class="control-label" for="ethereum-wallet-sendform-amount">'. __('Amount', 'ethereum-wallet') . '</label>
                <div class="input-group" style="margin-top: 8px">
                    <input style="cursor: text;" 
                        type="number"
                        value="" 
                        min="0"
                        step="0.000000000000000001"
                        id="ethereum-wallet-sendform-amount" 
                        name="ethereum-wallet-sendform-amount" 
                        class="form-control"
						lang="en-US">
                    <select style="max-width: 80px;"
                        class="custom-select form-control" 
                        id="ethereum-wallet-sendform-currency"
                        name="ethereum-wallet-sendform-currency" >
                        <option value="0x0000000000000000000000000000000000000001" selected="">'. __('ETH', 'ethereum-wallet') . '</option>
                        '.$ops.'
                    </select>
                </div>
            </div>
        </div>
    </div>
    <div class="row ethereum-wallet-sendform-content">
        <div class="col-12">
            <div class="form-group">
                '.wp_nonce_field( 'ethereum-wallet-send_form', 'ethereum-wallet-send-form-nonce', true, false ).'
                <input type="hidden" name="action" value="ethereum_wallet_send" />
                <button 
                    id="ethereum-wallet-send-button" 
                    name="ethereum-wallet-send-button" 
                    type="submit" 
                    value="'. __('Send', 'ethereum-wallet') . '" 
                    class="button btn btn-default float-right col-12 col-md-4">'. __('Send', 'ethereum-wallet') . '</button>
                <div id="ethereum-wallet-tx-in-progress-spinner" class="spinner float-right"></div>
            </div>
        </div>
    </div>
</div></div></form>';
    
    ETHEREUM_WALLET_enqueue_scripts_();
    wp_enqueue_script( 'jsQR' );

    return $js . str_replace("\n", " ", str_replace("\r", " ", str_replace("\t", " ", $js . $ret)));
}

add_shortcode( 'ethereum-wallet-sendform', 'ETHEREUM_WALLET_sendform_shortcode' );

function ETHEREUM_WALLET_sendform_action() {
    global $wp;
    
    if ( 'POST' !== strtoupper( $_SERVER['REQUEST_METHOD'] ) ) {
        return;
    }
    if ( empty( $_POST['action'] ) || 'ethereum_wallet_send' !== $_POST['action'] ) {
        return;
    }
    if (function_exists('wc_nocache_headers')) {
        wc_nocache_headers();
    } else {
        nocache_headers();
    }
    $nonce_value = '';
    if (isset($_REQUEST['ethereum-wallet-send-form-nonce'])) {
        $nonce_value = $_REQUEST['ethereum-wallet-send-form-nonce'];
    } else if (isset($_REQUEST['_wpnonce'])) {
        $nonce_value = $_REQUEST['_wpnonce'];
    }
    if ( ! wp_verify_nonce( $nonce_value, 'ethereum-wallet-send_form' ) ) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_sendform_action: bad nonce detected: " . $nonce_value);
        return;
    }
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    
    $accountAddress = get_user_meta( $user_id, 'user_ethereum_wallet_address', true);
    $privateKey = get_user_meta( $user_id, 'user_ethereum_wallet_key', true);

    // To address
    
    if (!isset( $_REQUEST['ethereum-wallet-sendform-to'] )) {
        ETHEREUM_WALLET_log("ethereum-wallet-sendform-to not set");
        return;
    }
    
    $to = sanitize_text_field($_REQUEST['ethereum-wallet-sendform-to']);
    if (empty( $to )) {
        ETHEREUM_WALLET_log("empty ethereum-wallet-sendform-to");
        return;
    }
    
    if (42 != strlen( $to )) {
        ETHEREUM_WALLET_log("strlen ethereum-wallet-sendform-to != 42: " . $to);
        return;
    }
    
    if ('0x' != substr( $to, 0, 2 )) {
        ETHEREUM_WALLET_log("startsWith ethereum-wallet-sendform-to != 0x: " . $to);
        return;
    }
    
    // Amount
    
    if (!isset( $_REQUEST['ethereum-wallet-sendform-amount'] )) {
        ETHEREUM_WALLET_log("ethereum-wallet-sendform-amount not set");
        return;
    }
    
    $amount = sanitize_text_field($_REQUEST['ethereum-wallet-sendform-amount']);
    if (empty( $amount )) {
        ETHEREUM_WALLET_log("empty ethereum-wallet-sendform-amount");
        return;
    }
    
    if (!is_numeric( $amount )) {
        ETHEREUM_WALLET_log("non-numeric ethereum-wallet-sendform-amount: " . $amount);
        return;
    }
        
    // Currency address
    
    if (!isset( $_REQUEST['ethereum-wallet-sendform-currency'] )) {
        ETHEREUM_WALLET_log("ethereum-wallet-sendform-currency not set");
        return;
    }
    
    $currency = sanitize_text_field($_REQUEST['ethereum-wallet-sendform-currency']);
    if (empty( $currency )) {
        ETHEREUM_WALLET_log("empty ethereum-wallet-sendform-currency");
        return;
    }
    
    if (42 != strlen( $currency )) {
        ETHEREUM_WALLET_log("strlen ethereum-wallet-sendform-currency != 42: " . $to);
        return;
    }
    
    if ('0x' != substr( $currency, 0, 2 )) {
        ETHEREUM_WALLET_log("startsWith ethereum-wallet-sendform-currency != 0x: " . $to);
        return;
    }
    
    if ("0x0000000000000000000000000000000000000001" === $currency) {
        // ETH
        $txhash = ETHEREUM_WALLET_send_ether($accountAddress, $to, $amount, $privateKey);
    }
    if (false !== $txhash) {
        update_user_meta( $user_id, 'user_ethereum_wallet_last_tx_currency', $currency);
        update_user_meta( $user_id, 'user_ethereum_wallet_last_tx_to', $to);
        update_user_meta( $user_id, 'user_ethereum_wallet_last_tx_value', $amount);
        update_user_meta( $user_id, 'user_ethereum_wallet_last_txhash', $txhash);
        update_user_meta( $user_id, 'user_ethereum_wallet_last_tx_hash', $txhash);
        update_user_meta( $user_id, 'user_ethereum_wallet_last_txtime', time());
?>
    <script>
        if ( window.history.replaceState ) {
            window.history.replaceState( null, null, window.location.href );
        }
    </script>
<?php
    }

}

add_action( 'wp_loaded', "ETHEREUM_WALLET_sendform_action", 20 );

function ETHEREUM_WALLET_account_management_create_shortcode( $attributes ) {

    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    
	$js = '';
	$ret = 
'<form method="post" action="" onsubmit="return ETHEREUM_WALLET_validate_account_management_create_form()"><div class="twbs"><div class="container-fluid ethereum-wallet-account-management-create-shortcode">
    <div class="row ethereum-wallet-account-management-create-content">
        <div class="col-12">
            <div class="form-group">
                <label class="control-label" for="ethereum-wallet-account-management-create-name">'.__('Account name', 'ethereum-wallet').'</label>
                <div class="input-group" style="margin-top: 8px">
                    <input type="text"
                           value="" 
                           placeholder="'.__('Input new account name', 'ethereum-wallet').'" 
                           id="ethereum-wallet-account-management-create-name" 
                           name="ethereum-wallet-account-management-create-name" 
                           class="form-control">
                </div>
            </div>
        </div>
    </div>
    <div class="row ethereum-wallet-account-management-create-content">
        <div class="col-12">
            <div class="form-group">
                '.wp_nonce_field( 'ethereum-wallet-account-management-create-send_form', 'ethereum-wallet-account-management-create-send-form-nonce', true, false ).'
                <input type="hidden" name="action" value="ethereum_wallet_account_management_create_send" />
                <button 
                    id="ethereum-wallet-account-management-create-send-button" 
                    name="ethereum-wallet-account-management-create-send-button" 
                    type="submit" 
                    value="'. __('Create', 'ethereum-wallet') . '" 
                    class="button btn btn-default float-right col-12 col-md-4">'. __('Create', 'ethereum-wallet') . '</button>
            </div>
        </div>
    </div>
</div></div></form>';
    
    ETHEREUM_WALLET_enqueue_scripts_();

    return $js . str_replace("\n", " ", str_replace("\r", " ", str_replace("\t", " ", $js . $ret)));
}

add_shortcode( 'ethereum-wallet-account-management-create', 'ETHEREUM_WALLET_account_management_create_shortcode' );

function ETHEREUM_WALLET_account_management_create_action() {
    global $wp;
    
    if ( 'POST' !== strtoupper( $_SERVER['REQUEST_METHOD'] ) ) {
        return;
    }
    if ( empty( $_POST['action'] ) || 'ethereum_wallet_account_management_create_send' !== $_POST['action'] ) {
        return;
    }
    if (function_exists('wc_nocache_headers')) {
        wc_nocache_headers();
    } else {
        nocache_headers();
    }
    $nonce_value = '';
    if (isset($_REQUEST['ethereum-wallet-account-management-create-send-form-nonce'])) {
        $nonce_value = $_REQUEST['ethereum-wallet-account-management-create-send-form-nonce'];
    } else if (isset($_REQUEST['_wpnonce'])) {
        $nonce_value = $_REQUEST['_wpnonce'];
    }
    if ( ! wp_verify_nonce( $nonce_value, 'ethereum-wallet-account-management-create-send_form' ) ) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_create_action: bad nonce detected: " . $nonce_value);
        return;
    }
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    
    if (isset( $_REQUEST['ethereum-wallet-account-management-create-name'] )) {
        $name = sanitize_text_field($_REQUEST['ethereum-wallet-account-management-create-name']);
        if (__( 'Default account', 'ethereum-wallet' ) == $name) {
            ETHEREUM_WALLET_log("ethereum-wallet-account-management-create-name an attempt to replace the default account is blocked: " . $name);
        }
        else if (!empty( $name )) {
            // create new account request
            ETHEREUM_WALLET_log("ethereum-wallet-account-management-create-name is set: " . $name);

            $accountsJSON = get_user_meta( $user_id, 'user_ethereum_wallet_accounts', true);
            if (empty($accountsJSON)) {
                $accountsJSON = '[]';
            }
            $accounts = json_decode($accountsJSON, true);
            list($ethAddressChkSum, $privateKeyHex) = ETHEREUM_WALLET_create_account();
            $accounts[] = [
                "name" => $name,
                "address" => $ethAddressChkSum,
                "key" => $privateKeyHex,
            ];
            update_user_meta( $user_id, 'user_ethereum_wallet_accounts', json_encode($accounts));
            
            // set new account as default
            update_user_meta( $user_id, 'user_ethereum_wallet_address', $ethAddressChkSum);
            update_user_meta( $user_id, 'user_ethereum_wallet_key', $privateKeyHex);
        }
    }
?>
    <script>
        if ( window.history.replaceState ) {
            window.history.replaceState( null, null, window.location.href );
        }
    </script>
<?php

}

add_action( 'wp_loaded', "ETHEREUM_WALLET_account_management_create_action", 20 );

function ETHEREUM_WALLET_account_management_import_shortcode( $attributes ) {

    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    
	$js = '';
	$ret = 
'<form method="post" action="" onsubmit="return ETHEREUM_WALLET_validate_account_management_import_form()"><div class="twbs"><div class="container-fluid ethereum-wallet-account-management-import-shortcode">
    <div class="row ethereum-wallet-account-management-import-content">
        <div class="col-12">
            <div class="form-group">
                <label class="control-label" for="ethereum-wallet-account-management-import-name">'.__('Account name', 'ethereum-wallet').'</label>
                <div class="input-group" style="margin-top: 8px">
                    <input type="text"
                        value="" 
                        placeholder="'.__('Input new account name', 'ethereum-wallet').'" 
                        id="ethereum-wallet-account-management-import-name" 
                        name="ethereum-wallet-account-management-import-name" 
                        class="form-control">
                </div>
            </div>
            <div class="form-group">
                <label class="control-label" for="ethereum-wallet-account-management-import-key">'.__('Private key', 'ethereum-wallet').'</label>
                <div class="input-group" style="margin-top: 8px">
                    <input type="text"
                        autocomplete="off"
                        value="" 
                        placeholder="'.__('Input your private key here', 'ethereum-wallet').'" 
                        id="ethereum-wallet-account-management-import-key" 
                        name="ethereum-wallet-account-management-import-key" 
                        class="form-control">
                </div>
            </div>
        </div>
    </div>
    <div class="row ethereum-wallet-account-management-import-content">
        <div class="col-12">
            <div class="form-group">
                '.wp_nonce_field( 'ethereum-wallet-account-management-import-send_form', 'ethereum-wallet-account-management-import-send-form-nonce', true, false ).'
                <input type="hidden" name="action" value="ethereum_wallet_account_management_import_send" />
                <button 
                    id="ethereum-wallet-account-management-import-send-button" 
                    name="ethereum-wallet-account-management-import-send-button" 
                    type="submit" 
                    value="'. __('Import', 'ethereum-wallet') . '" 
                    class="button btn btn-default float-right col-12 col-md-4">'. __('Import', 'ethereum-wallet') . '</button>
            </div>
        </div>
    </div>
</div></div></form>';
    
    ETHEREUM_WALLET_enqueue_scripts_();

    return $js . str_replace("\n", " ", str_replace("\r", " ", str_replace("\t", " ", $js . $ret)));
}

add_shortcode( 'ethereum-wallet-account-management-import', 'ETHEREUM_WALLET_account_management_import_shortcode' );

function ETHEREUM_WALLET_account_management_import_action() {
    global $wp;
    
    if ( 'POST' !== strtoupper( $_SERVER['REQUEST_METHOD'] ) ) {
        return;
    }
    if ( empty( $_POST['action'] ) || 'ethereum_wallet_account_management_import_send' !== $_POST['action'] ) {
        return;
    }
    if (function_exists('wc_nocache_headers')) {
        wc_nocache_headers();
    } else {
        nocache_headers();
    }
    $nonce_value = '';
    if (isset($_REQUEST['ethereum-wallet-account-management-import-send-form-nonce'])) {
        $nonce_value = $_REQUEST['ethereum-wallet-account-management-import-send-form-nonce'];
    } else if (isset($_REQUEST['_wpnonce'])) {
        $nonce_value = $_REQUEST['_wpnonce'];
    }
    if ( ! wp_verify_nonce( $nonce_value, 'ethereum-wallet-account-management-import-send_form' ) ) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_import_action: bad nonce detected: " . $nonce_value);
        return;
    }
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    
    if (isset( $_REQUEST['ethereum-wallet-account-management-import-name'] ) && 
        isset( $_REQUEST['ethereum-wallet-account-management-import-key'] )) {
        $name = sanitize_text_field($_REQUEST['ethereum-wallet-account-management-import-name']);
        $privateKeyHex = sanitize_text_field($_REQUEST['ethereum-wallet-account-management-import-key']);
        if (__( 'Default account', 'ethereum-wallet' ) == $name) {
            ETHEREUM_WALLET_log("ethereum-wallet-account-management-import-name an attempt to replace the default account is blocked: " . $name);
        }
        else if (!empty( $name ) && !empty( $privateKeyHex )) {
            // import new account request
            ETHEREUM_WALLET_log("ethereum-wallet-account-management-import-name is set: " . $name);
            $blnIsValid = false;
            try {
                $ecAdapter = \BitWasp\Bitcoin\Bitcoin::getEcAdapter();
                $privateKeyBuffer = \BitWasp\Buffertools\Buffer::hex($privateKeyHex);
                $blnIsValid = $ecAdapter->validatePrivateKey($privateKeyBuffer);
            } catch (Exception $ex) {
                ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_import_action: " . $ex->getMessage());
                ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_import_action: " . $ex->getTraceAsString());
            }
            if (!$blnIsValid) {
                ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_import_action: invalid private key");
            } else {
                $accountsJSON = get_user_meta( $user_id, 'user_ethereum_wallet_accounts', true);
                if (empty($accountsJSON)) {
                    $accountsJSON = '[]';
                }
                $accounts = json_decode($accountsJSON, true);
                $ethAddressChkSum = null;
                try {
                    $ethAddressChkSum = ETHEREUM_WALLET_address_from_key($privateKeyHex);
                } catch (Exception $ex) {
                    ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_import_action: " . $ex->getMessage());
                    ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_import_action: " . $ex->getTraceAsString());
                }
                if (!is_null($ethAddressChkSum)) {
                    $blnFound = false;
                    foreach ($accounts as $account) {
                        if ($ethAddressChkSum == $account["address"]) {
                            $blnFound = true;
                            break;
                        }
                    }
                    if (!$blnFound) {
                        $accounts[] = [
                            "name" => $name,
                            "address" => $ethAddressChkSum,
                            "key" => $privateKeyHex,
                            "imported" => true,
                        ];
                        update_user_meta( $user_id, 'user_ethereum_wallet_accounts', json_encode($accounts));

                        // set new account as default
                        update_user_meta( $user_id, 'user_ethereum_wallet_address', $ethAddressChkSum);
                        update_user_meta( $user_id, 'user_ethereum_wallet_key', $privateKeyHex);
                    }
                }
            }
        }
    }
?>
    <script>
        if ( window.history.replaceState ) {
            window.history.replaceState( null, null, window.location.href );
        }
    </script>
<?php

}

add_action( 'wp_loaded', "ETHEREUM_WALLET_account_management_import_action", 20 );

function ETHEREUM_WALLET_account_management_select_shortcode( $attributes ) {
//    global $ETHEREUM_WALLET_plugin_url_path;

//	$attributes = shortcode_atts( array(
//		'tokenname' => '',
//	), $attributes, 'ethereum-wallet' );
//	$options = stripslashes_deep( get_option( 'ethereum-wallet_options', array() ) );

    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }

    $ops = '';
    $accountsJSON = get_user_meta( $user_id, 'user_ethereum_wallet_accounts', true);
    if (empty($accountsJSON)) {
        $accountsJSON = '[]';
    }
    $accounts = json_decode($accountsJSON, true);
    if (!$accounts) {
        $name = __( 'Default account', 'ethereum-wallet' );
        $address = ETHEREUM_WALLET_get_wallet_address();
        $privateKey = get_user_meta( $user_id, 'user_ethereum_wallet_key', true);
        
        $accounts = [[
            "name" => $name,
            "address" => $address,
            "key" => $privateKey,
        ]];
        update_user_meta( $user_id, 'user_ethereum_wallet_accounts', json_encode($accounts));
    }
    $defaultAddress = ETHEREUM_WALLET_get_wallet_address();
    foreach ($accounts as $account) {
        $selected = '';
        if ($defaultAddress == $account["address"]) {
            $selected = ' selected';
        }
        $op = '<option value="' . $account["address"] . '"' . $selected . '>' . $account["name"] . ' - ' . $account["address"] . '</option>';
        $ops .= $op;
    }
    
	$js = '';
	$ret = 
'<form method="post" action="" onsubmit="return ETHEREUM_WALLET_validate_account_management_select_form()"><div class="twbs"><div class="container-fluid ethereum-wallet-account-management-select-shortcode">
    <div class="row ethereum-wallet-account-management-select-content">
        <div class="col-12">
            <div class="form-group">
                <label class="control-label" for="ethereum-wallet-account-management-select-default">'. __('Default account', 'ethereum-wallet') . '</label>
                <div class="input-group" style="margin-top: 8px">
                    <select 
                        class="custom-select form-control" 
                        id="ethereum-wallet-account-management-select-default"
                        name="ethereum-wallet-account-management-select-default" >
                        '.$ops.'
                    </select>
                </div>
            </div>
        </div>
    </div>
    <div class="row ethereum-wallet-account-management-select-content">
        <div class="col-12">
            <div class="form-group">
                '.wp_nonce_field( 'ethereum-wallet-account-management-select-send_form', 'ethereum-wallet-account-management-select-send-form-nonce', true, false ).'
                <input type="hidden" name="action" value="ethereum_wallet_account_management_select_send" />
                <button 
                    id="ethereum-wallet-account-management-select-send-button" 
                    name="ethereum-wallet-account-management-select-send-button" 
                    type="submit" 
                    value="'. __('Select', 'ethereum-wallet') . '" 
                    class="button btn btn-default float-right col-12 col-md-4">'. __('Select', 'ethereum-wallet') . '</button>
                <button 
                    id="ethereum-wallet-account-management-delete-send-button" 
                    name="ethereum-wallet-account-management-delete-send-button" 
                    type="submit" 
                    value="'. __('Remove', 'ethereum-wallet') . '" 
                    class="button btn btn-default float-right col-12 col-md-4">'. __('Remove', 'ethereum-wallet') . '</button>
            </div>
        </div>
    </div>
</div></div></form>';
    
    ETHEREUM_WALLET_enqueue_scripts_();

    return $js . str_replace("\n", " ", str_replace("\r", " ", str_replace("\t", " ", $js . $ret)));
}

add_shortcode( 'ethereum-wallet-account-management-select', 'ETHEREUM_WALLET_account_management_select_shortcode' );

function ETHEREUM_WALLET_account_management_select_action() {
    global $wp;
    
    if ( 'POST' !== strtoupper( $_SERVER['REQUEST_METHOD'] ) ) {
        return;
    }
    if ( empty( $_POST['action'] ) || 'ethereum_wallet_account_management_select_send' !== $_POST['action'] ) {
        return;
    }
    if (function_exists('wc_nocache_headers')) {
        wc_nocache_headers();
    } else {
        nocache_headers();
    }
    $nonce_value = '';
    if (isset($_REQUEST['ethereum-wallet-account-management-select-send-form-nonce'])) {
        $nonce_value = $_REQUEST['ethereum-wallet-account-management-select-send-form-nonce'];
    } else if (isset($_REQUEST['_wpnonce'])) {
        $nonce_value = $_REQUEST['_wpnonce'];
    }
    if ( ! wp_verify_nonce( $nonce_value, 'ethereum-wallet-account-management-select-send_form' ) ) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_select_action: bad nonce detected: " . $nonce_value);
        return;
    }
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }
    
    if (isset( $_REQUEST['ethereum-wallet-account-management-select-send-button'] )) {

        do {

            if (!isset( $_REQUEST['ethereum-wallet-account-management-select-default'] )) {
                break;
            }
            $newDefaultAddress = sanitize_text_field($_REQUEST['ethereum-wallet-account-management-select-default']);
            if (empty( $newDefaultAddress )) {
                ETHEREUM_WALLET_log("empty ethereum-wallet-account-management-select-default");
                break;
            }

            if (42 != strlen( $newDefaultAddress )) {
                ETHEREUM_WALLET_log("strlen ethereum-wallet-account-management-select-default != 42: " . $to);
                break;
            }

            if ('0x' != substr( $newDefaultAddress, 0, 2 )) {
                ETHEREUM_WALLET_log("startsWith ethereum-wallet-account-management-select-default != 0x: " . $to);
                break;
            }

            $defaultAddress = ETHEREUM_WALLET_get_wallet_address();
            if ($newDefaultAddress == $defaultAddress) {
                 ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_select_action: no difference with current default set: " . $defaultAddress);
                break;
            }

            $accountsJSON = get_user_meta( $user_id, 'user_ethereum_wallet_accounts', true);
            if (empty($accountsJSON)) {
                $accountsJSON = '[]';
            }
            $accounts = json_decode($accountsJSON, true);
            if (!$accounts) {
                break;
            }
            foreach ($accounts as $account) {
                if ($account["address"] == $newDefaultAddress) {
                    // set default account
                    ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_select_action: new default address set: " . $newDefaultAddress);
                    update_user_meta( $user_id, 'user_ethereum_wallet_address', $newDefaultAddress);
                    update_user_meta( $user_id, 'user_ethereum_wallet_key', $account["key"]);
                }
            }

        } while ( false );
    }
    else if (isset( $_REQUEST['ethereum-wallet-account-management-delete-send-button'] )) {

        do {

            if (!isset( $_REQUEST['ethereum-wallet-account-management-select-default'] )) {
                break;
            }
            $deleteAddress = sanitize_text_field($_REQUEST['ethereum-wallet-account-management-select-default']);
            ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_select_action: try to delete: " . $deleteAddress);
            if (empty( $deleteAddress )) {
                ETHEREUM_WALLET_log("empty ethereum-wallet-account-management-select-default");
                break;
            }

            if (42 != strlen( $deleteAddress )) {
                ETHEREUM_WALLET_log("strlen ethereum-wallet-account-management-select-default != 42: " . $to);
                break;
            }

            if ('0x' != substr( $deleteAddress, 0, 2 )) {
                ETHEREUM_WALLET_log("startsWith ethereum-wallet-account-management-select-default != 0x: " . $to);
                break;
            }

            $defaultAddress = get_user_meta( $user_id, 'user_ethereum_wallet_address', true);
            if ($deleteAddress == $defaultAddress) {
                 ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_select_action: can not delete current default address: " . $defaultAddress);
                break;
            }

            $accountsJSON = get_user_meta( $user_id, 'user_ethereum_wallet_accounts', true);
            if (empty($accountsJSON)) {
                $accountsJSON = '[]';
            }
            $accounts = json_decode($accountsJSON, true);
            if (!$accounts) {
                break;
            }
            $newAccounts = [];
            foreach ($accounts as $account) {
                if ($account["address"] == $deleteAddress && isset($account["imported"])) {
                    ETHEREUM_WALLET_log("ETHEREUM_WALLET_account_management_select_action: deleted address: " . $deleteAddress);
                    continue;
                }
                $newAccounts[] = $account;
            }
            update_user_meta( $user_id, 'user_ethereum_wallet_accounts', json_encode($newAccounts));
        } while ( false );
    }
?>
    <script>
        if ( window.history.replaceState ) {
            window.history.replaceState( null, null, window.location.href );
        }
    </script>
<?php

}

add_action( 'wp_loaded', "ETHEREUM_WALLET_account_management_select_action", 20 );

// TODO: wait for a configured number of blocks
/**
 * 
 * @param type $txhash
 * @return Integer confirmed: 1, unconfirmed: 0, failed: -1
 */
function ETHEREUM_WALLET_is_tx_confirmed($txhash) {
    $providerUrl = ETHEREUM_WALLET_getWeb3Endpoint();
    $requestManager = new \Web3\RequestManagers\HttpRequestManager($providerUrl, 10/* seconds */);
    $web3 = new \Web3\Web3(new \Web3\Providers\HttpProvider($requestManager));
    $eth = $web3->eth;
    $is_confirmed = false;
    $is_failed = false;
    $is_mined = false;
    $txBlockNumber = '';
    $savedTxInfo = null;
    $eth->getTransactionByHash($txhash, function ($err, $transaction) use(&$is_mined, &$is_failed, &$txBlockNumber, &$savedTxInfo) {
        if ($err !== null) {
            ETHEREUM_WALLET_log("Failed to getTransactionByHash: " . $err);
            $is_failed = true;
            return;
        }
        ETHEREUM_WALLET_log("transaction: " . print_r($transaction, true));
        $is_mined = property_exists($transaction, "blockHash") && !empty($transaction->blockHash) && '0x0000000000000000000000000000000000000000000000000000000000000000' != $transaction->blockHash;
        if ($is_mined) {
            $txBlockNumber = hexdec($transaction->blockNumber);
        }
        $savedTxInfo = $transaction;
    });
    if ($is_mined) {
        $eth->getTransactionReceipt($txhash, function ($err, $receipt) use(&$is_confirmed, &$is_failed, &$savedTxInfo) {
            if ($err !== null) {
                ETHEREUM_WALLET_log("Failed to getTransactionReceipt: " . $err);
                return;
            }
            ETHEREUM_WALLET_log("transaction receipt: " . print_r($receipt, true));
            if (isset($receipt->status)) {
                if ($receipt->status === "0x1") {
                    $is_confirmed = true;
                } else if ($receipt->status === "0x0") {
                    if ($receipt->gasUsed > $savedTxInfo->gas) {
                        $is_failed = true;
                        ETHEREUM_WALLET_log("getTransactionReceipt($txhash): we ran out of gas, not confirmed!");
                    } else {
                        $is_failed = true;
                        ETHEREUM_WALLET_log("getTransactionReceipt($txhash): bad tx status, not confirmed!");
                    }
                } else {
                    // unknown status. pre-Byzantium
                    if ($receipt->gasUsed >= $savedTxInfo->gas) {
                        $is_failed = true;
                        ETHEREUM_WALLET_log("getTransactionReceipt($txhash): we ran out of gas, not confirmed!");
                    } else {
                        $is_confirmed = true;
                    }
                }
            } else {
                // unknown status. pre-Byzantium
                if ($receipt->gasUsed >= $savedTxInfo->gas) {
                    $is_failed = true;
                    ETHEREUM_WALLET_log("getTransactionReceipt($txhash): we ran out of gas, not confirmed!");
                } else {
                    $is_confirmed = true;
                }
            }
        });
        if ($is_confirmed) {
            $blockNumber = null;
            $eth->blockNumber(function ($err, $lastBlockNumber) use(&$blockNumber) {
                if ($err !== null) {
                    ETHEREUM_WALLET_log("Failed to get blockNumber: " . $err);
                    return;
                }
                ETHEREUM_WALLET_log("lastBlockNumber: " . $lastBlockNumber->toString());
                $blockNumber = intval($lastBlockNumber->toString());
            });
            if (null !== $blockNumber && '' !== $txBlockNumber) {
                // https://www.reddit.com/r/ethereum/comments/4eplsv/how_many_confirms_is_considered_safe_in_ethereum/d229xie/
                $safeBlockCount = 12; // TODO: add admin setting
                $is_confirmed = $is_confirmed && ($blockNumber - $txBlockNumber >= $safeBlockCount);
            }
            ETHEREUM_WALLET_log("is_confirmed($txhash in block $txBlockNumber): " . $is_confirmed);
        }
    }
    return $is_confirmed ? 1 : ($is_failed ? -1 : 0);
}

function ETHEREUM_WALLET_history_shortcode( $attributes ) {
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        return;
    }

	$attributes = shortcode_atts( array(
		'direction' => '',
	), $attributes, 'ethereum-wallet' );

    // The displayed tx direction: in/out/inout
    $direction = ! empty( $attributes['direction'] ) ? $attributes['direction'] : 'inout';
    if (!in_array($direction, array('in', 'out', 'inout'))) {
        $direction = 'inout';
    }
	$js = '';
	$ret = 
'<div class="twbs"><div class="container-fluid ethereum-wallet-history-shortcode">
    <div class="row ethereum-wallet-history-table-wrapper">
        <div class="col-12">
            <div class="table-responsive">
                <table class="table table-striped table-condensed ethereum-wallet-history-table ethereum-wallet-history-table-direction-'.$direction.'">
                    <thead>
                        <tr>
                            <th>'.__('#', 'ethereum-wallet').'</th>
                            <th>'.__('Amount', 'ethereum-wallet').'</th>
                            <th>'.__('Date', 'ethereum-wallet').'</th>
                            <th>'.__('Tx', 'ethereum-wallet').'</th>
                        </tr>
                    </thead>
                    <tbody>
                    </tbody>
                </table>
            </div>
        </div>
    </div>
</div></div>';

    ETHEREUM_WALLET_enqueue_scripts_();
    wp_enqueue_script( 'data-tables' );

    return $js . str_replace("\n", " ", str_replace("\r", " ", str_replace("\t", " ", $js . $ret)));
}

add_shortcode( 'ethereum-wallet-history', 'ETHEREUM_WALLET_history_shortcode' );

function ETHEREUM_WALLET_enqueue_scripts_() {
    wp_enqueue_style( 'ethereum-wallet' );
    wp_enqueue_script( 'ethereum-wallet' );
}

function ETHEREUM_WALLET_stylesheet() {
	global $ETHEREUM_WALLET_plugin_url_path;
	
    $deps = array('font-awesome', 'bootstrap-ethereum-wallet', 'bootstrap-affix-ethereum-wallet', 'data-tables');
    $min = defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ? '' : '.min';
    if( ( ! wp_style_is( 'font-awesome', 'queue' ) ) && ( ! wp_style_is( 'font-awesome', 'done' ) ) ) {
        wp_dequeue_style('font-awesome');
        wp_deregister_style('font-awesome');
        wp_register_style(
            'font-awesome', 
            $ETHEREUM_WALLET_plugin_url_path . "/css/font-awesome{$min}.css", array(), '4.7.0'
        );
    }
    if ( !wp_style_is( 'bootstrap-ethereum-wallet', 'queue' ) && !wp_style_is( 'bootstrap-ethereum-wallet', 'done' ) ) {
        wp_dequeue_style('bootstrap-ethereum-wallet');
        wp_deregister_style('bootstrap-ethereum-wallet');
        wp_register_style(
            'bootstrap-ethereum-wallet', 
            $ETHEREUM_WALLET_plugin_url_path . "/css/bootstrap-ns{$min}.css", array(), '4.0.0'
        );
    }
    if ( !wp_style_is( 'bootstrap-affix-ethereum-wallet', 'queue' ) && !wp_style_is( 'bootstrap-affix-ethereum-wallet', 'done' ) ) {
        wp_dequeue_style('bootstrap-affix-ethereum-wallet');
        wp_deregister_style('bootstrap-affix-ethereum-wallet');
        wp_register_style(
            'bootstrap-affix-ethereum-wallet', 
            $ETHEREUM_WALLET_plugin_url_path . "/css/affix.css", array(), '3.3.7'
        );
    }
    if( ( ! wp_style_is( 'data-tables', 'queue' ) ) && ( ! wp_style_is( 'data-tables', 'done' ) ) ) {
        wp_dequeue_style('data-tables');
        wp_deregister_style('data-tables');
        wp_register_style(
            'data-tables', 
            $ETHEREUM_WALLET_plugin_url_path . "/css/jquery.dataTables{$min}.css", array(), '1.10.19'
        );
    }
    if( ( ! wp_style_is( 'ethereum-wallet', 'queue' ) ) && ( ! wp_style_is( 'ethereum-wallet', 'done' ) ) ) {
        wp_dequeue_style('ethereum-wallet');
        wp_deregister_style('ethereum-wallet');
        wp_register_style( 
            'ethereum-wallet', 
            $ETHEREUM_WALLET_plugin_url_path . '/ethereum-wallet.css', 
            $deps, 
            '1.14.0' 
        );
    }
}

add_action( 'wp_enqueue_scripts', 'ETHEREUM_WALLET_stylesheet', 20 );

function ETHEREUM_WALLET_get_wallet_address() {
    $accountAddress = '';
    $user_id = get_current_user_id();
    if ($user_id > 0) {
        $accountAddress = get_user_meta( $user_id, 'user_ethereum_wallet_address', true);
        if (empty($accountAddress)) {
            ETHEREUM_WALLET_user_registration( $user_id );
            $accountAddress = get_user_meta( $user_id, 'user_ethereum_wallet_address', true);
        }
    }
    return $accountAddress;
}

function ETHEREUM_WALLET_get_wallet_name() {
    $accountName = '';
    $user_id = get_current_user_id();
    if ($user_id <= 0) {
        return $accountName;
    }
    $accountAddress = get_user_meta( $user_id, 'user_ethereum_wallet_address', true);
    $accountsJSON = get_user_meta( $user_id, 'user_ethereum_wallet_accounts', true);
    if (empty($accountsJSON)) {
        $accountsJSON = '[]';
    }
    $accounts = json_decode($accountsJSON, true);
    if ($accounts) {
        foreach($accounts as $account) {
            if ($account["address"] != $accountAddress) {
                continue;
            }
            $accountName = $account['name'];
        }
    }
    if (empty($accountName)) {
        $accountName = __( 'Default account', 'ethereum-wallet' );
    }
    return $accountName;
}

function ETHEREUM_WALLET_get_last_tx_hash_time() {
    $lastTxHash = '';
    $lastTxTime = '';

    $user_id = get_current_user_id();
    if ($user_id > 0) {
        $lastTxHash = get_user_meta( $user_id, 'user_ethereum_wallet_last_txhash', true);
        $lastTxTime = get_user_meta( $user_id, 'user_ethereum_wallet_last_txtime', true);
        
        if (!empty($lastTxHash) && 0 != ETHEREUM_WALLET_is_tx_confirmed($lastTxHash)) {
            delete_user_meta( $user_id, 'user_ethereum_wallet_last_txhash', $lastTxHash);
            delete_user_meta( $user_id, 'user_ethereum_wallet_last_txtime', $lastTxTime);
            $lastTxHash = '';
            $lastTxTime = '';
        }

        if( empty( $lastTxTime ) ) {
            $lastTxTime = '';
        } else {
            $lastTxTime = time() - intval($lastTxTime);
        }
    }
    return [$lastTxHash, $lastTxTime];
}

function ETHEREUM_WALLET_enqueue_script() {
	global $ETHEREUM_WALLET_plugin_url_path;
	
    $min = defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ? '' : '.min';

    if( ( ! wp_script_is( 'web3', 'queue' ) ) && ( ! wp_script_is( 'web3', 'done' ) ) ) {
        wp_dequeue_script('web3');
        wp_deregister_script('web3');
        wp_register_script( 
            'web3', 
            //$ETHEREUM_WALLET_plugin_url_path . "/js/web3{$min}.js", array( 'jquery' ), '0.20.6'
            $ETHEREUM_WALLET_plugin_url_path . "/js/web3.min.js", array( 'jquery' ), '0.20.6'
        );
    }

    if( ( ! wp_script_is( 'data-tables', 'queue' ) ) && ( ! wp_script_is( 'data-tables', 'done' ) ) ) {
        wp_dequeue_script('data-tables');
        wp_deregister_script('data-tables');
        wp_register_script( 
            'data-tables', 
            //$ETHEREUM_WALLET_plugin_url_path . "/js/jquery.dataTables{$min}.js", array( 'jquery' ), '1.10.19'
            $ETHEREUM_WALLET_plugin_url_path . "/js/jquery.dataTables.min.js", array( 'jquery' ), '1.10.19'
        );
    }

    if( ( ! wp_script_is( 'popper', 'queue' ) ) && ( ! wp_script_is( 'popper', 'done' ) ) ) {
        wp_dequeue_script('popper');
        wp_deregister_script('popper');
        wp_register_script(
            'popper', 
//            $ETHEREUM_WALLET_plugin_url_path . "/js/popper{$min}.js", array('jquery'), '1.14.6'
            $ETHEREUM_WALLET_plugin_url_path . "/js/popper.min.js", array('jquery'), '1.14.6'
        );
    }

    if( ( ! wp_script_is( 'ethereum-wallet-bootstrap', 'queue' ) ) && ( ! wp_script_is( 'ethereum-wallet-bootstrap', 'done' ) ) ) {
        wp_dequeue_script('ethereum-wallet-bootstrap');
        wp_deregister_script('ethereum-wallet-bootstrap');
        wp_register_script(
            'ethereum-wallet-bootstrap', 
//            $ETHEREUM_WALLET_plugin_url_path . "/js/bootstrap{$min}.js", array('jquery', 'popper'), '4.0.0'
            $ETHEREUM_WALLET_plugin_url_path . "/js/bootstrap.min.js", array('jquery', 'popper'), '4.0.0'
        );
    }

    if( ( ! wp_script_is( 'ethereum-wallet-bootstrap-affix', 'queue' ) ) && ( ! wp_script_is( 'ethereum-wallet-bootstrap-affix', 'done' ) ) ) {
        wp_dequeue_script('ethereum-wallet-bootstrap-affix');
        wp_deregister_script('ethereum-wallet-bootstrap-affix');
        wp_register_script(
            'ethereum-wallet-bootstrap-affix', 
            $ETHEREUM_WALLET_plugin_url_path . "/js/affix.js", array('ethereum-wallet-bootstrap'), '3.3.7'
        );
    }

    if( ( ! wp_script_is( 'qrcode', 'queue' ) ) && ( ! wp_script_is( 'qrcode', 'done' ) ) ) {
        wp_dequeue_script('qrcode');
        wp_deregister_script('qrcode');
        wp_register_script(
            'qrcode', 
//            $ETHEREUM_WALLET_plugin_url_path . "/js/qrcode{$min}.js", array('ethereum-wallet-bootstrap-affix'), '2009'
            $ETHEREUM_WALLET_plugin_url_path . "/js/qrcode.min.js", array('ethereum-wallet-bootstrap-affix'), '2009'
        );
    }

    if( ( ! wp_script_is( 'jquery.qrcode', 'queue' ) ) && ( ! wp_script_is( 'jquery.qrcode', 'done' ) ) ) {
        wp_dequeue_script('jquery.qrcode');
        wp_deregister_script('jquery.qrcode');
        wp_register_script(
            'jquery.qrcode', 
//            $ETHEREUM_WALLET_plugin_url_path . "/js/jquery.qrcode{$min}.js", array('jquery', 'qrcode'), '1.0'
            $ETHEREUM_WALLET_plugin_url_path . "/js/jquery.qrcode.min.js", array('jquery', 'qrcode'), '1.0'
        );
    }

    // https://github.com/cozmo/jsQR
    if( ( ! wp_script_is( 'jsQR', 'queue' ) ) && ( ! wp_script_is( 'jsQR', 'done' ) ) ) {
        wp_dequeue_script('jsQR');
        wp_deregister_script('jsQR');
        wp_register_script(
            'jsQR', 
//            $ETHEREUM_WALLET_plugin_url_path . "/js/jsQR{$min}.js", array('ethereum-wallet-bootstrap-affix'), '807b07357a35a2d16d8006bcc5426ce558d6b904'
           $ETHEREUM_WALLET_plugin_url_path . "/js/jsQR.min.js", array('ethereum-wallet-bootstrap-affix'), '807b07357a35a2d16d8006bcc5426ce558d6b904'
        );
    }

    if( ( ! wp_script_is( 'clipboard', 'queue' ) ) && ( ! wp_script_is( 'clipboard', 'done' ) ) ) {
        wp_dequeue_script('clipboard');
        wp_deregister_script('clipboard');
        wp_register_script(
            'clipboard', 
//            $ETHEREUM_WALLET_plugin_url_path . "/js/clipboard{$min}.js", array(), '2.0.4'
           $ETHEREUM_WALLET_plugin_url_path . "/js/clipboard.min.js", array(), '2.0.4'
        );
    }

    if( ( ! wp_script_is( 'ethereum-wallet', 'queue' ) ) && ( ! wp_script_is( 'ethereum-wallet', 'done' ) ) ) {
        wp_dequeue_script('ethereum-wallet');
        wp_deregister_script('ethereum-wallet');
        wp_register_script( 
            'ethereum-wallet', 
            $ETHEREUM_WALLET_plugin_url_path . "/ethereum-wallet{$min}.js",
            array( 'jquery', 'web3', 'ethereum-wallet-bootstrap-affix' ),
            '1.14.0'
        );
    }

	$options = stripslashes_deep( get_option( 'ethereum-wallet_options', array() ) );
	
	$etherscanApiKey = ! empty( $options['etherscanApiKey'] ) ? esc_attr( $options['etherscanApiKey'] ) : 
        '';
	$blockchain_network = ! empty( $options['blockchain_network'] ) ? esc_attr( $options['blockchain_network'] ) : 
        'mainnet';
	$gaslimit = ! empty( $options['gaslimit'] ) ? esc_attr( $options['gaslimit'] ) : "200000";
    $gasprice = ETHEREUM_WALLET_get_gas_price_wei();

    $accountAddress = '';
    $lastTxHash = '';
    $lastTxTime = '';
    $lastTxTo = '';
    $lastTxValue = '';
    $lastTxCurrency = '';
    $tokens_json = '[]';
    $user_id = get_current_user_id();
    if ($user_id > 0) {
        $accountAddress = ETHEREUM_WALLET_get_wallet_address();

        $lastTxHash = get_user_meta( $user_id, 'user_ethereum_wallet_last_txhash', true);
        $lastTxTime = get_user_meta( $user_id, 'user_ethereum_wallet_last_txtime', true);
        $lastTxTo = get_user_meta( $user_id, 'user_ethereum_wallet_last_tx_to', true);
        $lastTxValue = get_user_meta( $user_id, 'user_ethereum_wallet_last_tx_value', true);
        $lastTxCurrency = get_user_meta( $user_id, 'user_ethereum_wallet_last_tx_currency', true);
        
        if (!empty($lastTxHash) && 0 != ETHEREUM_WALLET_is_tx_confirmed($lastTxHash)) {
            delete_user_meta( $user_id, 'user_ethereum_wallet_last_txhash', $lastTxHash);
            delete_user_meta( $user_id, 'user_ethereum_wallet_last_txtime', $lastTxTime);
            $lastTxHash = '';
            $lastTxTime = '';
        }

        if( empty( $lastTxTime ) ) {
            $lastTxTime = '';
        } else {
            $lastTxTime = time() - intval($lastTxTime);
        }
    }
    
    wp_localize_script(
        'ethereum-wallet', 'ethereumWallet', [
            // variables
            'user_wallet_address' => esc_html($accountAddress),
            'user_wallet_last_txhash' => esc_html($lastTxHash),
            'user_wallet_last_txtime' => esc_html($lastTxTime),
            'user_wallet_last_tx_to' => esc_html($lastTxTo),
            'user_wallet_last_tx_value' => esc_html($lastTxValue),
            'user_wallet_last_tx_currency' => esc_html($lastTxCurrency),
            'tokens' => esc_html($tokens_json),
            'web3Endpoint' => esc_html(ETHEREUM_WALLET_getWeb3Endpoint()),
            'blockchain_network' => esc_html($blockchain_network),
            'gasLimit' => esc_html($gaslimit),
            'gasPrice' => esc_html($gasprice),
            'localePath' => esc_html($ETHEREUM_WALLET_plugin_url_path . "/i18n/" . get_locale() . ".json"),
            // translations
            'str_copied_msg' => __('Copied to clipboard', 'ethereum-wallet'),
            'str_insufficient_eth_balance_msg' => __('Insufficient Ether balance for tx fee payment.', 'ethereum-wallet'),
            'str_unknown_token_symbol_msg' => __('Unknown', 'ethereum-wallet'),
            'str_tx_pending_msg' => __('Pending', 'ethereum-wallet'),
            'str_date_recently_msg' => __('recently', 'ethereum-wallet'),
            'str_date_days_fmt_msg' => __('%1$s days', 'ethereum-wallet'),
            'str_date_hours_fmt_msg' => __('%1$s hours', 'ethereum-wallet'),
            'str_date_minutes_fmt_msg' => __('%1$s minutes', 'ethereum-wallet'),
        ]
    );
}

add_action( 'wp_enqueue_scripts', 'ETHEREUM_WALLET_enqueue_script' );

/**
 * Admin Options
 */

if ( is_admin() ) {
	include_once $ETHEREUM_WALLET_plugin_dir . '/ethereum-wallet.admin.php';
}

function ETHEREUM_WALLET_add_menu_link() {
	$page = add_options_page(
		__( 'Ethereum Wallet Settings', 'ethereum-wallet' ),
		__( 'Ethereum Wallet', 'ethereum-wallet' ),
		'manage_options',
		'ethereum-wallet',
		'ETHEREUM_WALLET_options_page'
	);
}

add_filter( 'admin_menu', 'ETHEREUM_WALLET_add_menu_link' );

// Place in Option List on Settings > Plugins page 
function ETHEREUM_WALLET_actlinks( $links, $file ) {
	// Static so we don't call plugin_basename on every plugin row.
	static $this_plugin;
	
	if ( ! $this_plugin ) {
		$this_plugin = plugin_basename( __FILE__ );
	}
	
	if ( $file == $this_plugin ) {
		$settings_link = '<a href="options-general.php?page=ethereum-wallet">' . __( 'Settings' ) . '</a>';
		array_unshift( $links, $settings_link ); // before other links
	}
	
	return $links;
}

add_filter( 'plugin_action_links', 'ETHEREUM_WALLET_actlinks', 10, 2 );

function ETHEREUM_WALLET_get_default_gas_price_gwei() {
    global $ETHEREUM_WALLET_options;
    $gasPriceMaxGwei = doubleval(isset($ETHEREUM_WALLET_options['gas_price']) ? $ETHEREUM_WALLET_options['gas_price'] : '41');
    return array('tm' => time(), 'gas_price' => $gasPriceMaxGwei);
}
    
function ETHEREUM_WALLET_query_gas_price_gwei() {    
    $apiEndpoint = "https://www.etherchain.org/api/gasPriceOracle";
    $response = wp_remote_get( $apiEndpoint, array('sslverify' => false) );
    if( is_wp_error( $response ) ) {
        ETHEREUM_WALLET_log("Error in gasPriceOracle response: ", $response);
        return ETHEREUM_WALLET_get_default_gas_price_gwei();
    }
    
    $http_code = wp_remote_retrieve_response_code( $response );
    if (200 != $http_code) {
        ETHEREUM_WALLET_log("Bad response code in gasPriceOracle response: ", $http_code);
        return ETHEREUM_WALLET_get_default_gas_price_gwei();
    }
    
    $body = wp_remote_retrieve_body( $response );
    if (!$body) {
        ETHEREUM_WALLET_log("empty body in gasPriceOracle response");
        return ETHEREUM_WALLET_get_default_gas_price_gwei();
    }
    
    $j = json_decode($body, true);
    if (!isset($j["fast"])) {
        ETHEREUM_WALLET_log("no fast field in gasPriceOracle response");
        return ETHEREUM_WALLET_get_default_gas_price_gwei();
    }
    
    $gasPriceGwei = $j["fast"];
    $cache_gas_price = array('tm' => time(), 'gas_price' => $gasPriceGwei);

    if ( get_option('ethereumicoio_cache_gas_price') ) {
        update_option('ethereumicoio_cache_gas_price', $cache_gas_price);
    } else {
        $deprecated='';
        $autoload='no';
        add_option('ethereumicoio_cache_gas_price', $cache_gas_price, $deprecated, $autoload);
    }
    return $cache_gas_price;
}

function ETHEREUM_WALLET_get_gas_price_wei() {
    // Get all existing Cryptocurrency Product options
    $cache_gas_price_gwei = get_option( 'ethereumicoio_cache_gas_price', array() );
    if (!$cache_gas_price_gwei) {
        $cache_gas_price_gwei = ETHEREUM_WALLET_query_gas_price_gwei();
    }
    
    $tm_diff = time() - intval($cache_gas_price_gwei['tm']);
    // TODO: admin setting
    $timeout = 10 * 60; // seconds
    if ($tm_diff > $timeout) {
        $cache_gas_price_gwei = ETHEREUM_WALLET_query_gas_price_gwei();
    }
    
    $gasPriceGwei = doubleval($cache_gas_price_gwei['gas_price']);
    $gasPriceMaxGwei = doubleval(ETHEREUM_WALLET_get_default_gas_price_gwei()['gas_price']);
    
    if ($gasPriceMaxGwei < $gasPriceGwei) {
        $gasPriceGwei = $gasPriceMaxGwei;
    }
    $gasPriceWei = 1000000000 * $gasPriceGwei; // gwei -> wei
    return intval($gasPriceWei);
}

function ETHEREUM_WALLET_send_transaction($to, $eth_value_wei, $data) {
    global $ETHEREUM_WALLET_options;
    
    $chainId = ETHEREUM_WALLET_getChainId();
    if (null === $chainId) {
        ETHEREUM_WALLET_log("chainId is null");
        return false;
    }
    
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_send_transaction: no user");
        return false;
    }
    $from = get_user_meta( $user_id, 'user_ethereum_wallet_address', true);
    if (empty($from)) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_send_transaction: empty from address");
        return false;
    }
    
    $providerUrl = ETHEREUM_WALLET_getWeb3Endpoint();
    $gasLimit = 21000;
    if (!empty($data)) {
        $gasLimit = intval(isset($ETHEREUM_WALLET_options['gas_limit']) ? $ETHEREUM_WALLET_options['gas_limit'] : '200000');
    }
    $gasPrice = ETHEREUM_WALLET_get_gas_price_wei();
    
    $eth_value_wei = new phpseclib\Math\BigInteger($eth_value_wei);
    $eth_value_with_fee_wei = $eth_value_wei->add((new phpseclib\Math\BigInteger($gasLimit))->multiply(new phpseclib\Math\BigInteger($gasPrice)));
    
    // 1. check deposit
    $eth_balance = ETHEREUM_WALLET_getBalanceEth($providerUrl, $from);
    if (null === $eth_balance) {
        ETHEREUM_WALLET_log("eth_balance is null");
        return false;
    }
    if ($eth_balance->compare($eth_value_with_fee_wei) < 0) {
        $eth_balance_str = $eth_balance->toString();
        $eth_value_with_fee_wei_str = $eth_value_with_fee_wei->toString();
        ETHEREUM_WALLET_log("Insufficient funds: eth_balance_wei($eth_balance_str) < eth_value_with_fee_wei($eth_value_with_fee_wei_str)");
        return false;
    }
    $requestManager = new \Web3\RequestManagers\HttpRequestManager($providerUrl, 10/* seconds */);
    $web3 = new \Web3\Web3(new \Web3\Providers\HttpProvider($requestManager));
    $eth = $web3->eth;
    $nonce = 0;
    $eth->getTransactionCount($from, function ($err, $transactionCount) use(&$nonce) {
        if ($err !== null) {
            ETHEREUM_WALLET_log("Failed to getTransactionCount: " . $err);
            $nonce = null;
            return;
        }
        $nonce = intval($transactionCount->toString());
    });
    if (null === $nonce) {
        return false;
    }

    $nonce = \BitWasp\Buffertools\Buffer::int($nonce);
    $gasPrice = \BitWasp\Buffertools\Buffer::int($gasPrice);
    $gasLimit = \BitWasp\Buffertools\Buffer::int($gasLimit);
    $value = $eth_value_wei->toHex();
    //ETHEREUM_WALLET_log("value: " . $value);

    $transactionParamsArray = [
        'nonce' => '0x' . $nonce->getHex(),
        'to' => strtolower($to),
        'gas' => '0x' . $gasLimit->getHex(),
        'gasPrice' => '0x' . $gasPrice->getHex(),
        'value' => '0x' . $value,
        'chainId' => $chainId,
        'data' => $data
    ];
    $transaction = new \Web3p\EthereumTx\Transaction($transactionParamsArray);
    $privateKey = get_user_meta( $user_id, 'user_ethereum_wallet_key', true);
    if (empty($privateKey)) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_send_transaction: empty key");
        return false;
    }
    $signedTransaction = "0x" . $transaction->sign($privateKey);
    //ETHEREUM_WALLET_log("transaction: " . print_r($transactionParamsArray, true));
    //ETHEREUM_WALLET_log("signedTransaction: " . $signedTransaction);

    $txHash = null;
    $eth->sendRawTransaction((string)$signedTransaction, function ($err, $transaction) use(&$txHash) {
        if ($err !== null) {
            ETHEREUM_WALLET_log("Failed to sendRawTransaction: " . $err);
            return;
        }
        $txHash = $transaction;
    });
    
    if (null === $txHash) {
        ETHEREUM_WALLET_log("Failed to sendRawTransaction: txHash == null");
        return false;
    }
    
    ETHEREUM_WALLET_log("txHash: " . $txHash);
    
    if (false !== $txHash) {
        update_user_meta( $user_id, 'user_ethereum_wallet_last_txhash', $txHash);
        update_user_meta( $user_id, 'user_ethereum_wallet_last_tx_hash', $txHash);
        update_user_meta( $user_id, 'user_ethereum_wallet_last_txtime', time());
        update_user_meta( $user_id, 'user_ethereum_wallet_last_tx_to', $transactionParamsArray['to']);
        update_user_meta( $user_id, 'user_ethereum_wallet_last_tx_value', $transactionParamsArray['value']);
    }
    return $txHash;
}

function ETHEREUM_WALLET_sign_transaction($to, $eth_value_wei, $data, $gasLimit, $gasPrice, $nonce) {
    global $ETHEREUM_WALLET_options;
    
    $chainId = ETHEREUM_WALLET_getChainId();
    if (null === $chainId) {
        ETHEREUM_WALLET_log("chainId is null");
        return false;
    }
    
    $user_id = get_current_user_id();
    if ( $user_id <= 0 ) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_send_transaction: no user");
        return false;
    }
    $from = get_user_meta( $user_id, 'user_ethereum_wallet_address', true);
    if (empty($from)) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_send_transaction: empty from address");
        return false;
    }
    
    $providerUrl = ETHEREUM_WALLET_getWeb3Endpoint();
    
    $eth_value_wei = new phpseclib\Math\BigInteger($eth_value_wei, 16);
    $eth_value_with_fee_wei = $eth_value_wei->add(
        (new phpseclib\Math\BigInteger(str_replace('0x', '', $gasLimit), 16))
            ->multiply(new phpseclib\Math\BigInteger(str_replace('0x', '', $gasPrice), 16))
    );
    
    // 1. check deposit
    $eth_balance = ETHEREUM_WALLET_getBalanceEth($providerUrl, $from);
    if (null === $eth_balance) {
        ETHEREUM_WALLET_log("eth_balance is null");
        return false;
    }
    if ($eth_balance->compare($eth_value_with_fee_wei) < 0) {
        $eth_balance_str = $eth_balance->toString();
        $eth_value_with_fee_wei_str = $eth_value_with_fee_wei->toString();
        ETHEREUM_WALLET_log("Insufficient funds: eth_balance_wei($eth_balance_str) < eth_value_with_fee_wei($eth_value_with_fee_wei_str)");
        return false;
    }
    $requestManager = new \Web3\RequestManagers\HttpRequestManager($providerUrl, 10/* seconds */);
    $web3 = new \Web3\Web3(new \Web3\Providers\HttpProvider($requestManager));
    $eth = $web3->eth;

    $value = $eth_value_wei->toHex();
    //ETHEREUM_WALLET_log("value: " . $value);

    $transactionParamsArray = [
        'nonce' => $nonce,
        'to' => strtolower($to),
        'gas' => $gasLimit,
        'gasPrice' => $gasPrice,
        'value' => '0x' . $value,
        'chainId' => $chainId,
        'data' => $data
    ];
    $transaction = new \Web3p\EthereumTx\Transaction($transactionParamsArray);
    $privateKey = get_user_meta( $user_id, 'user_ethereum_wallet_key', true);
    if (empty($privateKey)) {
        ETHEREUM_WALLET_log("ETHEREUM_WALLET_send_transaction: empty key");
        return false;
    }
    $signedTransaction = "0x" . $transaction->sign($privateKey);
    ETHEREUM_WALLET_log("transaction: " . print_r($transactionParamsArray, true));
    //ETHEREUM_WALLET_log("signedTransaction: " . $signedTransaction);

    return $signedTransaction;
}

function ETHEREUM_WALLET_send_ether($from, $to, $eth_value, $privateKey) {
    global $ETHEREUM_WALLET_options;
    
    $chainId = ETHEREUM_WALLET_getChainId();
    if (null === $chainId) {
        ETHEREUM_WALLET_log("chainId is null");
        return false;
    }
    
    $providerUrl = ETHEREUM_WALLET_getWeb3Endpoint();
    $gasLimit = 21000;
    $gasPrice = ETHEREUM_WALLET_get_gas_price_wei();
    
    $eth_value_wei = _ETHEREUM_WALLET_double_int_multiply($eth_value, pow(10, 18));
    $eth_value_with_fee_wei = $eth_value_wei->add((new phpseclib\Math\BigInteger($gasLimit))->multiply(new phpseclib\Math\BigInteger($gasPrice)));
    
    // 1. check deposit
    $eth_balance = ETHEREUM_WALLET_getBalanceEth($providerUrl, $from);
    if (null === $eth_balance) {
        ETHEREUM_WALLET_log("eth_balance is null");
        return false;
    }
    if ($eth_balance->compare($eth_value_with_fee_wei) < 0) {
        $eth_balance_str = $eth_balance->toString();
        $eth_value_with_fee_wei_str = $eth_value_with_fee_wei->toString();
        ETHEREUM_WALLET_log("Insufficient funds: eth_balance_wei($eth_balance_str) < eth_value_with_fee_wei($eth_value_with_fee_wei_str)");
        return false;
    }
    $requestManager = new \Web3\RequestManagers\HttpRequestManager($providerUrl, 10/* seconds */);
    $web3 = new \Web3\Web3(new \Web3\Providers\HttpProvider($requestManager));
    $eth = $web3->eth;
    $nonce = 0;
    $eth->getTransactionCount($from, function ($err, $transactionCount) use(&$nonce) {
        if ($err !== null) {
            ETHEREUM_WALLET_log("Failed to getTransactionCount: " . $err);
            $nonce = null;
            return;
        }
        $nonce = intval($transactionCount->toString());
    });
    if (null === $nonce) {
        return false;
    }
    $data = '';

    $nonce = \BitWasp\Buffertools\Buffer::int($nonce);
    $gasPrice = \BitWasp\Buffertools\Buffer::int($gasPrice);
    $gasLimit = \BitWasp\Buffertools\Buffer::int($gasLimit);
    $value = $eth_value_wei->toHex();
    //ETHEREUM_WALLET_log("value: " . $value);

    $transactionParamsArray = [
        'nonce' => '0x' . $nonce->getHex(),
        'to' => strtolower($to),
        'gas' => '0x' . $gasLimit->getHex(),
        'gasPrice' => '0x' . $gasPrice->getHex(),
        'value' => '0x' . $value,
        'chainId' => $chainId,
        'data' => ''
    ];
    $transaction = new \Web3p\EthereumTx\Transaction($transactionParamsArray);
    $signedTransaction = "0x" . $transaction->sign($privateKey);
    //ETHEREUM_WALLET_log("transaction: " . print_r($transactionParamsArray, true));
    ETHEREUM_WALLET_log("signedTransaction: " . $signedTransaction);

    $txHash = null;
    $eth->sendRawTransaction((string)$signedTransaction, function ($err, $transaction) use(&$txHash) {
        if ($err !== null) {
            ETHEREUM_WALLET_log("Failed to sendRawTransaction: " . $err);
            return;
        }
        $txHash = $transaction;
    });
    
    if (null === $txHash) {
        return false;
    }
    
    ETHEREUM_WALLET_log("txHash: " . $txHash);
    
    return $txHash;
}

function ETHEREUM_WALLET_getChainId() {
    $blockchainNetwork = ETHEREUM_WALLET_getBlockchainNetwork();
    if (empty($blockchainNetwork)) {
        $blockchainNetwork = 'mainnet';
    }
    if ($blockchainNetwork === 'mainnet') {
        return 1;
    }
    if ($blockchainNetwork === 'ropsten') {
        return 3;
    }
    if ($blockchainNetwork === 'rinkeby') {
        return 4;
    }
    ETHEREUM_WALLET_log("Bad blockchain_network setting:" . $blockchainNetwork);
    return null;
}

function ETHEREUM_WALLET_get_txhash_path($txHash) {
    $blockchainNetwork = ETHEREUM_WALLET_getBlockchainNetwork();
    $txHashPath = '';
    switch ($blockchainNetwork) {
        case 'mainnet':
            $txHashPath = 'https://etherscan.io/tx/' . $txHash;
            break;
        case 'ropsten':
            $txHashPath = 'https://ropsten.etherscan.io/tx/' . $txHash;
            break;
        case 'rinkeby':
            $txHashPath = 'https://rinkeby.etherscan.io/tx/' . $txHash;
            break;
        default:
            break;
    }
    return $txHashPath;
}

function _ETHEREUM_WALLET_double_int_multiply($dval, $ival) {
    $dval = doubleval($dval);
    $ival = intval($ival);
    $dv1 = floor($dval);
    $ret = new phpseclib\Math\BigInteger(intval($dv1));
    $ret = $ret->multiply(new phpseclib\Math\BigInteger($ival));
    if ($dv1 === $dval) {
        return $ret;
    }
    $dv2 = $dval - $dv1;
    $iv1 = intval($dv2 * $ival);
    $ret = $ret->add(new phpseclib\Math\BigInteger($iv1));
    return $ret;
}
        
} // PHP version
